/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#ifndef SCHEME_H
#define SCHEME_H

#include <stdio.h>
#include <setjmp.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#ifndef SCHEME_DIRECT_EMBEDDED
#define SCHEME_DIRECT_EMBEDDED 1
#endif

#ifdef __cplusplus
extern "C" 
{
#endif

typedef short Scheme_Type;

typedef struct Scheme_Bucket
{
  Scheme_Type type;
  char *key;
  void *val;
} Scheme_Bucket;

typedef struct Scheme_Hash_Table
{
  Scheme_Type type;
  int size, count, step;
  Scheme_Bucket **buckets;
  char by_symbol, by_address, has_constants, forever;
} Scheme_Hash_Table;

/* Hash tablekey types, used with scheme_hash_table */
enum {
  SCHEME_hash_string,
  SCHEME_hash_stringptr,
  SCHEME_hash_ptr,
  SCHEME_hash_symbol
};

#define SCHEME_HAS_SET_BANG 1
#define SCHEME_HAS_ANNOTATIONS 2

typedef struct Scheme_Basic_Env 
{
  short num_bindings;
  char autounbox;
  char flags; /* used for expanding/compilaing */
  struct Scheme_Env *next;
} Scheme_Basic_Env;

#define SCHEME_AUTO_UNBOX 1
#define SCHEME_ANNOTATING 2

typedef struct Scheme_Env
{
  struct Scheme_Basic_Env basic;
  struct Scheme_Object *values[1];
} Scheme_Env;

typedef struct Scheme_Compile_Info
{
  int base_let_depth;
  int max_let_depth;
  short can_return_closure;
  short can_optimize_constants;
} Scheme_Compile_Info;

typedef struct Scheme_Object *
(Scheme_Prim)(int argc, struct Scheme_Object *argv[]);

typedef struct Scheme_Object *
(Scheme_Closed_Prim)(void *d, int argc, struct Scheme_Object *argv[]);

typedef struct Scheme_Object *
(Scheme_Syntax)(struct Scheme_Object *form, struct Scheme_Env *env,
		struct Scheme_Compile_Info *rec);

typedef struct Scheme_Object *
(Scheme_Syntax_Expander)(struct Scheme_Object *form, struct Scheme_Env *env,
			 int depth);

typedef struct Scheme_Object *
(Scheme_Syntax_Compilation)(struct Scheme_Object *data, 
			    struct Scheme_Env *env);

typedef struct Scheme_Object *
(Scheme_Method_Prim)(struct Scheme_Object *o, 
		     int argc, struct Scheme_Object *argv[]);

typedef struct Scheme_Object
{
  Scheme_Type type; /* Anything that starts with a type field
		       can be a Scheme_Object */
  union
    {
      struct { char *string_val; int tag_val; } str_val;
      struct { void *ptr1, *ptr2; } two_ptr_val;
      struct { int int1; int int2; } two_int_val;
      struct { void *ptr; int pint; } ptr_int_val;
      struct {
	Scheme_Closed_Prim *f;
	void *d; } clsd_prim_val;
      struct { struct Scheme_Object *car, *cdr; } pair_val;
      struct { int size; struct Scheme_Object **els; } vector_val;
      struct { struct Scheme_Env *env; struct Scheme_Object *code; } closure_val;
      struct {
	Scheme_Syntax *compile;
	Scheme_Syntax_Expander *expand;
      } syntax_val;
      struct Scheme_Debugging_Info *debug_val;
    } u;
} Scheme_Object;

typedef struct {
  Scheme_Type type;
  union {
    char char_val;
    void *ptr_value;
    long int_val;
    void *ptr_val;
    Scheme_Prim *prim_val;
  } u;
} Scheme_Small_Object;  

typedef struct {
  Scheme_Type type;
  double double_val;
} Scheme_Double;

typedef struct {
  Scheme_Type type;
  Scheme_Prim *prim_val;
  short mina, maxa;
} Scheme_Primitive_Proc;

typedef struct {
  Scheme_Type type;
  Scheme_Closed_Prim *prim_val;
  void *data;
  short mina, maxa;
} Scheme_Closed_Primitive_Proc;

typedef struct Scheme_Debugging_Info {
  Scheme_Object *src;
} Scheme_Debugging_Info;

typedef struct Scheme_Trigger {
  Scheme_Type type;
  int value;
} Scheme_Trigger;

typedef struct Scheme_Symbol {
  Scheme_Type type;
  char s[1];
} Scheme_Symbol;

typedef struct Scheme_Input_Port
{
  Scheme_Type type;
  Scheme_Object *sub_type;
  void *port_data;
  int (*getc_fun) (struct Scheme_Input_Port *port);
  int (*char_ready_fun) (struct Scheme_Input_Port *port);
  void (*close_fun) (struct Scheme_Input_Port *port);
  void (*need_wakeup_fun)(struct Scheme_Input_Port *, void *);
  char *name;
  char *ungotten;
  int ungotten_count, ungotten_allocated;
  long position, lineNumber, charsSinceNewline;
  int eoffound;
  int closed;
} Scheme_Input_Port;

typedef struct Scheme_Output_Port
{
  Scheme_Type type;
  Scheme_Object *sub_type;
  void *port_data;
  void (*write_string_fun)(char *str, long len, struct Scheme_Output_Port *);
  void (*close_fun) (struct Scheme_Output_Port *);
  int closed;
} Scheme_Output_Port;

typedef struct {
  Scheme_Type type;
  struct Scheme_Object *sclass;
  void *primdata; /* For your use */
  short primflag; /* For your use */
  short inited;
} Scheme_Class_Object;

/* Like setjmp & longjmp, but you can jmp to a deeper stack position */
/* Intialize a Scheme_Jumpup_Buf record before using it */
typedef struct Scheme_Jumpup_Buf {
  void *stack_from, *stack_copy;
  long stack_size, stack_max_size;
  struct Scheme_Jumpup_Buf *cont;
  jmp_buf buf;
} Scheme_Jumpup_Buf;

typedef struct Scheme_Process {
  Scheme_Type type;

  jmp_buf error_buf;
  int jumping_to_continuation;

  Scheme_Env *env;
  Scheme_Hash_Table *globals;

  Scheme_Object *in_port, *stdin_port, *write_port;
  Scheme_Object *out_port, *stdout_port, *read_port;
  Scheme_Object *err_port, *stderr_port, *readerr_port;

  Scheme_Object *error_handler_proc;

  int (*check_for_break)(void);
  Scheme_Object *user_break_poll_proc;
  int enable_break;

  Scheme_Object *exit_handler;

  int engine_weight;

  void *stack_start, *stack_end;
  Scheme_Jumpup_Buf jmpup_buf;
  
  void *cc_start;
  long *cc_ok;
  struct Scheme_Dynamic_Wind *dw;

  struct Scheme_Process *next;

  int running, stack_overflow;

  int block_descriptor;
  Scheme_Object *blocker; /* trigger or port */
  int ran_some;

  Scheme_Object *(*evaluator)(Scheme_Object *);

  struct Scheme_Env_Stack *env_stack;

  struct Scheme_Overflow *overflow;
  jmp_buf overflow_buf;

  Scheme_Object *eval_handler;
} Scheme_Process;

/* Type readers & writers for compiled code data */
typedef Scheme_Object *(*Scheme_Type_Reader)(Scheme_Object *list);
typedef Scheme_Object *(*Scheme_Type_Writer)(Scheme_Object *obj);

/* This file defines all the built-in types */
#include "../src/stypes.h"



#ifndef DOS_MEMORY
#define FAST_NUMBERS /* Force fast numbers */
#endif

/* Value-access macros */
#ifdef FAST_NUMBERS
#define SCHEME_TYPE(obj)     (SCHEME_INTP(obj)?(Scheme_Type)scheme_integer_type:(obj)->type)
#define _SCHEME_TYPE(obj) ((obj)->type) /* unsafe version */
#else
#define SCHEME_TYPE(obj)     ((obj)->type)
#define _SCHEME_TYPE SCHEME_TYPE
#endif


#define SCHEME_CHAR_VAL(obj) (((Scheme_Small_Object *)(obj))->u.char_val)
#ifdef FAST_NUMBERS
#define SCHEME_INT_VAL(obj)  (((long)obj)>>1)
#else
#define SCHEME_INT_VAL(obj)  (((Scheme_Small_Object *)(obj))->u.int_val)
#endif
#define SCHEME_DBL_VAL(obj)  (((Scheme_Double *)obj)->double_val)
#define SCHEME_NUM_VAL(obj) (SCHEME_INTP(obj)?SCHEME_INT_VAL(obj):SCHEME_DBL_VAL(obj))
#define SCHEME_STR_VAL(obj)  ((obj)->u.str_val.string_val)
#define SCHEME_STRTAG_VAL(obj)  ((obj)->u.str_val.tag_val)
#define SCHEME_STRLEN_VAL(obj)  ((obj)->u.str_val.tag_val)
#define SCHEME_SYM_VAL(obj)  (((Scheme_Symbol *)(obj))->s)
#define SCHEME_TSYM_VAL(obj)  (SCHEME_SYM_VAL(SCHEME_PTR_VAL(obj)))
#define SCHEME_PTR_VAL(obj)  (((Scheme_Small_Object *)(obj))->u.ptr_val)
#define SCHEME_PTR1_VAL(obj) ((obj)->u.two_ptr_val.ptr1)
#define SCHEME_PTR2_VAL(obj) ((obj)->u.two_ptr_val.ptr2)
#define SCHEME_IPTR_VAL(obj) ((obj)->u.ptr_int_val.ptr)
#define SCHEME_INT1_VAL(obj) ((obj)->u.two_int_val.int1)
#define SCHEME_INT2_VAL(obj) ((obj)->u.two_int_val.int2)
#define SCHEME_PINT_VAL(obj) ((obj)->u.ptr_int_val.pint)
#define SCHEME_SYNTAX(obj)   ((obj)->u.syntax_val.compile)
#define SCHEME_SYNTAX_EXP(obj)   ((obj)->u.syntax_val.expand)
#define SCHEME_PRIM(obj)     (((Scheme_Primitive_Proc *)(obj))->prim_val)
#define SCHEME_CLSD_PRIM(obj) (((Scheme_Closed_Primitive_Proc *)(obj))->prim_val)
#define SCHEME_CLSD_PRIM_DATA(obj) (((Scheme_Closed_Primitive_Proc *)(obj))->data)
#define SCHEME_CAR(obj)      ((obj)->u.pair_val.car)
#define SCHEME_CDR(obj)      ((obj)->u.pair_val.cdr)
#define SCHEME_VEC_SIZE(obj) ((obj)->u.vector_val.size)
#define SCHEME_VEC_ELS(obj)  ((obj)->u.vector_val.els)
#define SCHEME_CLOS_ENV(obj) ((obj)->u.closure_val.env)
#define SCHEME_CLOS_CODE(obj) ((obj)->u.closure_val.code)
#define SCHEME_DEBUG(obj)    ((obj)->u.debug_val)
#define SCHEME_OBJ_CLASS(obj) ((Scheme_Object *)((Scheme_Class_Object *)(obj))->sclass)
#define SCHEME_OBJ_DATA(obj)  (((Scheme_Class_Object *)(obj))->primdata)
#define SCHEME_OBJ_FLAG(obj)  (((Scheme_Class_Object *)(obj))->primflag)
#define SCHEME_INPORT_VAL(obj) (((Scheme_Input_Port)(obj))->port_data)
#define SCHEME_OUTPORT_VAL(obj) (((Scheme_Output_Port)(obj))->port_data)
#define SCHEME_VAR_BUCKET(obj) ((Scheme_Bucket *)(obj))

#define SCHEME_ASSERT(expr,msg) ((expr) ? 1 : (scheme_signal_error(msg), 0))

#if !SCHEME_DIRECT_EMBEDDED
#define scheme_current_process (*scheme_current_process_ptr)
#endif

#define scheme_error_buf (scheme_current_process->error_buf)
#define scheme_jumping_to_continuation (scheme_current_process->jumping_to_continuation)
#define scheme_env (scheme_current_process->env)
#define scheme_stdin_port (scheme_current_process->stdin_port)
#define scheme_stdout_port (scheme_current_process->stdout_port)
#define scheme_stderr_port (scheme_current_process->stderr_port)

#define scheme_enable_break (scheme_current_process->enable_break)
#define scheme_check_for_break (scheme_current_process->check_for_break)

#define scheme_evaluator (scheme_current_process->evaluator)
#define scheme_eval_handler (scheme_current_process->eval_handler)

#define scheme_setjmpup(b, s) scheme_setjmpup_relative(b, s, NULL)

#if SCHEME_DIRECT_EMBEDDED
extern Scheme_Object *scheme_tail_call_waiting, *scheme_eval_waiting;
#define scheme_tail_eval(obj, env) \
 (scheme_eval_wait_expr = obj, scheme_eval_wait_env = env, scheme_eval_waiting)
#endif

/* Allocation */
#define scheme_alloc_object() \
   ((Scheme_Object *) scheme_malloc (sizeof (Scheme_Object)))
#define scheme_alloc_small_object() \
   ((Scheme_Object *) scheme_malloc (sizeof (Scheme_Small_Object)))
#define scheme_alloc_stubborn_object() \
   ((Scheme_Object *) scheme_malloc_stubborn (sizeof (Scheme_Object)))
#define scheme_alloc_stubborn_small_object() \
   ((Scheme_Object *) scheme_malloc_stubborn (sizeof (Scheme_Small_Object)))
#define scheme_alloc_eternal_object() \
   ((Scheme_Object *) scheme_malloc_eternal (sizeof (Scheme_Object)))
#define scheme_alloc_eternal_small_object() \
   ((Scheme_Object *) scheme_malloc_eternal (sizeof (Scheme_Small_Object)))

#ifdef SCHEME_NO_GC
void *scheme_malloc(size_t size);
#define scheme_malloc_atomic scheme_malloc
#define scheme_malloc_stubborn scheme_malloc
#define scheme_malloc_uncollectable scheme_malloc
#else
#define scheme_malloc GC_malloc
#define scheme_malloc_atomic GC_malloc_atomic
#define scheme_malloc_stubborn GC_malloc_stubborn
#define scheme_malloc_uncollectable GC_malloc_uncollectable
#endif

#ifdef FAST_NUMBERS
#define scheme_make_integer(i) ((Scheme_Object *)((((long)i) << 1) | 0x1))
#else
#define scheme_make_integer scheme_make_integer_value
#endif

#define scheme_new_frame(n) scheme_new_special_frame(n, 0)
#define scheme_extend_env(f, e) (f->basic.next = e, f)
#define scheme_next_frame(e) ((e)->basic.next)
#define scheme_settable_frame(f, s) ((f)->basic.has_set_bang = (s))
#define scheme_get_frame_settable(f) ((f)->basic.has_set_bang)
#define scheme_get_binding(f, n) ((f)->values[n])

#if SCHEME_DIRECT_EMBEDDED

/* These flags must be set before MzScheme is started: */
extern int scheme_case_sensitive; /* Defaults to 0 */
extern int scheme_constant_builtins; /* Defaults to 0 */
extern int scheme_no_keywords; /* Defaults to 0 */
extern int scheme_allow_set_undefined; /* Defaults to 0 */
extern int scheme_escape_continuations_only; /* Defaults to 0 */

extern Scheme_Process *scheme_current_process, *scheme_first_process;

/* Set these global hooks: */
extern void (*scheme_exit)(int v);
extern void (*scheme_console_printf)(char *str, ...);
extern void (*scheme_sleep)(int seconds, void *fds);
extern void (*scheme_notify_multithread)(int on);
extern void (*scheme_wakeup_on_input)(void *fds);

extern Scheme_Object *(*scheme_make_stdin)(void);
extern Scheme_Object *(*scheme_make_stdout)(void);
extern Scheme_Object *(*scheme_make_stderr)(void);

/* Initialization */
Scheme_Env *scheme_basic_env(void);

/* All functions & global constants prototyped here */
#include "../src/schemef.h"

#else

/* Constants and function prototypes in a record: */
#include "../src/schemex.h"

extern Scheme_Extension_Table *scheme_extension_table;

/* Macro mapping names to record access */
#include "../src/schemexm.h"

#endif

#ifdef WIN32S_HACK    /* See comment in setjmpup.c */
#define scheme_longjmp(b, v) \
    { jmp_buf hack; setjmp(hack); (b)->j_excep = hack->j_excep; \
      longjmp(b, v); }
#else
#define scheme_longjmp(b, v) longjmp(b, v)
#endif
#define scheme_setjmp(b) setjmp(b)

#ifndef DOS_MEMORY
#define SAME_PTR(a, b) ((a) == (b))
#define NOT_SAME_PTR(a, b) ((a) != (b))
#else
/* DOS isn't actually supported now: No garbage collector */
extern int scheme_same_pointer(void *a, void *b);
extern int scheme_diff_pointer(void *a, void *b);
#define SAME_PTR(a, b) scheme_same_pointer((void *)(a), (void *)(b))
#define NOT_SAME_PTR(a, b) scheme_diff_pointer((void *)(a), (void *)(b))
#endif

#define SAME_OBJ SAME_PTR
#define NOT_SAME_OBJ NOT_SAME_PTR

#define SAME_TYPE(a, b) ((Scheme_Type)(a) == (Scheme_Type)(b))
#define NOT_SAME_TYPE(a, b) ((Scheme_Type)(a) != (Scheme_Type)(b))

/* convenience macros */
#define SCHEME_CHARP(obj)    SAME_TYPE(SCHEME_TYPE(obj), scheme_char_type)
#ifdef FAST_NUMBERS
#define SCHEME_INTP(obj)     (((long)obj) & 0x1)
#else
#define SCHEME_INTP(obj)     SAME_TYPE(SCHEME_TYPE(obj), scheme_integer_type)
#endif
#define SCHEME_DBLP(obj)     SAME_TYPE(SCHEME_TYPE(obj), scheme_double_type)
#define SCHEME_BIGNUMP(obj)     SAME_TYPE(SCHEME_TYPE(obj), scheme_bignum_type)
#define SCHEME_NUMBERP(obj)  (SCHEME_INTP(obj) || SCHEME_DBLP(obj) || SCHEME_BIGNUMP(obj))
#define SCHEME_STRINGP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_string_type)
#define SCHEME_SYMBOLP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_symbol_type)
#define SCHEME_TSYMBOLP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_type_symbol_type)
#define SCHEME_BOOLP(obj)    (SAME_OBJ(obj, scheme_true) || SAME_OBJ(obj, scheme_false))
#define SCHEME_FALSEP(obj)     SAME_OBJ((obj), scheme_false)
#define SCHEME_TRUEP(obj)     (!SCHEME_FALSEP(obj))
#define SCHEME_SYNTAXP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_compiled_syntax_type)
#define SCHEME_PRIMP(obj)    SAME_TYPE(SCHEME_TYPE(obj), scheme_prim_type)
#define SCHEME_CLSD_PRIMP(obj)    SAME_TYPE(SCHEME_TYPE(obj), scheme_closed_prim_type)
#define SCHEME_CONTP(obj)    SAME_TYPE(SCHEME_TYPE(obj), scheme_cont_type)
#define SCHEME_ECONTP(obj)    SAME_TYPE(SCHEME_TYPE(obj), scheme_escaping_cont_type)
#define SCHEME_NULLP(obj)    SAME_OBJ(obj, scheme_null)
#define SCHEME_PAIRP(obj)    SAME_TYPE(SCHEME_TYPE(obj), scheme_pair_type)
#define SCHEME_LISTP(obj)    (SCHEME_NULLP(obj) || SCHEME_PAIRP(obj))
#define SCHEME_BOXP(obj)     SAME_TYPE(SCHEME_TYPE(obj), scheme_box_type)
#define SCHEME_HASHTP(obj) SAME_TYPE(SCHEME_TYPE(obj),scheme_hash_table_type)
#define SCHEME_VECTORP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_vector_type)
#define SCHEME_STRUCT_PROCP(obj) SAME_TYPE(SCHEME_TYPE(obj), scheme_struct_proc_type)
#define SCHEME_CLOSUREP(obj) (SAME_TYPE(SCHEME_TYPE(obj), scheme_closure_type) || SAME_TYPE(SCHEME_TYPE(obj), scheme_compiled_closure_type) || SAME_TYPE(SCHEME_TYPE(obj), scheme_case_closure_type))
#define SCHEME_PROCP(obj)    (SCHEME_PRIMP(obj) || SCHEME_CLSD_PRIMP(obj) || SCHEME_CLOSUREP(obj) || SCHEME_CONTP(obj) || SCHEME_ECONTP(obj) || SCHEME_STRUCT_PROCP(obj))
#define SCHEME_INPORTP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_input_port_type)
#define SCHEME_OUTPORTP(obj) SAME_TYPE(SCHEME_TYPE(obj), scheme_output_port_type)
#define SCHEME_EOFP(obj)     SAME_OBJ((obj), scheme_eof)
#define SCHEME_PROMP(obj)    SAME_TYPE(SCHEME_TYPE(obj), scheme_promise_type)
#define SCHEME_DEBUGP(obj)   SAME_TYPE(SCHEME_TYPE(obj), scheme_debug_handle_type)
#define SCHEME_PROCESSP(obj)   SAME_TYPE(SCHEME_TYPE(obj), scheme_process_type)
#define SCHEME_TRIGGERP(obj)   SAME_TYPE(SCHEME_TYPE(obj), scheme_trigger_type)
#define SCHEME_OBJP(obj)     SAME_TYPE(SCHEME_TYPE(obj), scheme_object_type)
#define SCHEME_CLASSP(obj)   SAME_TYPE(SCHEME_TYPE(obj), scheme_class_type)
#define SCHEME_VOIDP(obj)     SAME_OBJ((obj), scheme_void)
#define SCHEME_DIVARP(obj) SAME_TYPE(SCHEME_TYPE(obj), scheme_delayed_ivar_type)

/* other */
#define SCHEME_CADR(obj)     (SCHEME_CAR (SCHEME_CDR (obj)))
#define SCHEME_CAAR(obj)     (SCHEME_CAR (SCHEME_CAR (obj)))
#define SCHEME_CDDR(obj)     (SCHEME_CDR (SCHEME_CDR (obj)))
#define SCHEME_IPORT_NAME(obj) (((Scheme_Input_Port *)obj)->name)

#ifdef __cplusplus
};
#endif

#endif /* ! SCHEME_H */

