/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <ctype.h>

/* globals */

/* locals */
static Scheme_Object *char_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_lt (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_gt (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_lt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_gt_eq (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_eq_ci (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_lt_ci (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_gt_ci (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_lt_eq_ci (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_gt_eq_ci (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_alphabetic (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_numeric (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_whitespace (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_upper_case (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_lower_case (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_to_integer (int argc, Scheme_Object *argv[]);
static Scheme_Object *integer_to_char (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_upcase (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_downcase (int argc, Scheme_Object *argv[]);

void scheme_init_char (Scheme_Env *env)
{
  scheme_add_global_constant("char?", scheme_make_prim_w_arity(char_p, 1, 1), env);
  scheme_add_global_constant("char=?", scheme_make_prim_w_arity(char_eq, 2, 2), env);
  scheme_add_global_constant("char<?", scheme_make_prim_w_arity(char_lt, 2, 2), env);
  scheme_add_global_constant("char>?", scheme_make_prim_w_arity(char_gt, 2, 2), env);
  scheme_add_global_constant("char<=?", scheme_make_prim_w_arity(char_lt_eq, 2, 2), env);
  scheme_add_global_constant("char>=?", scheme_make_prim_w_arity(char_gt_eq, 2, 2), env);
  scheme_add_global_constant("char-ci=?", scheme_make_prim_w_arity(char_eq_ci, 2, 2), env);
  scheme_add_global_constant("char-ci<?", scheme_make_prim_w_arity(char_lt_ci, 2, 2), env);
  scheme_add_global_constant("char-ci>?", scheme_make_prim_w_arity(char_gt_ci, 2, 2), env);
  scheme_add_global_constant("char-ci<=?", scheme_make_prim_w_arity(char_lt_eq_ci, 2, 2), env);
  scheme_add_global_constant("char-ci>=?", scheme_make_prim_w_arity(char_gt_eq_ci, 2, 2), env);
  scheme_add_global_constant("char-alphabetic?", scheme_make_prim_w_arity(char_alphabetic, 1, 1), env);
  scheme_add_global_constant("char-numeric?", scheme_make_prim_w_arity(char_numeric, 1, 1), env);
  scheme_add_global_constant("char-whitespace?", scheme_make_prim_w_arity(char_whitespace, 1, 1), env);
  scheme_add_global_constant("char-upper-case?", scheme_make_prim_w_arity(char_upper_case, 1, 1), env);
  scheme_add_global_constant("char-lower-case?", scheme_make_prim_w_arity(char_lower_case, 1, 1), env);
  scheme_add_global_constant("char->integer", scheme_make_prim_w_arity(char_to_integer, 1, 1), env);
  scheme_add_global_constant("integer->char", scheme_make_prim_w_arity(integer_to_char, 1, 1), env);
  scheme_add_global_constant("char-upcase", scheme_make_prim_w_arity(char_upcase, 1, 1), env);
  scheme_add_global_constant("char-downcase", scheme_make_prim_w_arity(char_downcase, 1, 1), env);
}

Scheme_Object *
scheme_make_char (char ch)
{
  Scheme_Object *sc;

  sc = scheme_alloc_stubborn_small_object ();
  sc->type = scheme_char_type;
  SCHEME_CHAR_VAL (sc) = ch;
  scheme_end_stubborn_change((void *)sc);
  return (sc);
}

/* locals */

static Scheme_Object *
char_p (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("char?", 1, 1, argc, argv);

  return (SCHEME_CHARP(argv[0]) ? scheme_true : scheme_false);
}

#define CHAR_BIN_CHECK(name) \
 if ((argc != 2) || !SCHEME_CHARP(argv[0]) || !SCHEME_CHARP(argv[1])) \
   char_bin_error(name, argc, argv)

#define CHAR_UN_CHECK(name) \
 if ((argc != 2) || !SCHEME_CHARP(argv[0])) \
   char_un_error(name, argc, argv)

static void char_bin_error(char *name, int argc, Scheme_Object *argv[])
{
  if (argc != 2)
    scheme_wrong_count(name, 2, 2, argc, argv);
  if (!SCHEME_CHARP(argv[0]))
    scheme_wrong_type(name, "character", 0, argc, argv);
  if (!SCHEME_CHARP(argv[1]))
    scheme_wrong_type(name, "character", 1, argc, argv);
}

static void char_un_error(char *name, int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count(name, 1, 1, argc, argv);
  if (!SCHEME_CHARP(argv[0]))
    scheme_wrong_type(name, "character", 0, argc, argv);
}

static Scheme_Object *
char_eq (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char=?");

  return ((SCHEME_CHAR_VAL(argv[0]) == SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_lt (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char<?");

  return ((SCHEME_CHAR_VAL(argv[0]) < SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_gt (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char>?");

  return ((SCHEME_CHAR_VAL(argv[0]) > SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_lt_eq (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char<=?");

  return ((SCHEME_CHAR_VAL(argv[0]) <= SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_gt_eq (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char>=?");

  return ((SCHEME_CHAR_VAL(argv[0]) >= SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_eq_ci (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char-ci=?");

  return (toupper(SCHEME_CHAR_VAL(argv[0])) == toupper(SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_lt_ci (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char-ci<?");

  return (toupper(SCHEME_CHAR_VAL(argv[0])) < toupper(SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_gt_ci (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char-ci>?");

  return (toupper(SCHEME_CHAR_VAL(argv[0])) > toupper(SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_lt_eq_ci (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char-ci<=?");

  return (toupper(SCHEME_CHAR_VAL(argv[0])) <= toupper(SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_gt_eq_ci (int argc, Scheme_Object *argv[])
{
  CHAR_BIN_CHECK("char-ci>=?");

  return (toupper(SCHEME_CHAR_VAL(argv[0])) >= toupper(SCHEME_CHAR_VAL(argv[1]))
	  ? scheme_true : scheme_false);
}

static Scheme_Object *
char_alphabetic (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char-alphabetic?");

  return (isalpha (SCHEME_CHAR_VAL (argv[0])) ? scheme_true : scheme_false);
}

static Scheme_Object *
char_numeric (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char-numeric?");

  return (isdigit (SCHEME_CHAR_VAL (argv[0])) ? scheme_true : scheme_false);
}

static Scheme_Object *
char_whitespace (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char-whitespace?");

  return (isspace (SCHEME_CHAR_VAL (argv[0])) ? scheme_true : scheme_false);
}

static Scheme_Object *
char_upper_case (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char-upper-case?");

  return (isupper (SCHEME_CHAR_VAL (argv[0])) ? scheme_true : scheme_false);
}

static Scheme_Object *
char_lower_case (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char-lower-case?");

  return (islower (SCHEME_CHAR_VAL (argv[0])) ? scheme_true : scheme_false);
}

static Scheme_Object *
char_to_integer (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char->integer");

  return (scheme_make_integer (SCHEME_CHAR_VAL (argv[0])));
}

static Scheme_Object *
integer_to_char (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("integer->char", 1, 1, argc, argv);
  if (!SCHEME_INTP(argv[0]))
    scheme_wrong_type("integer->char", "integer", 0, argc, argv);

  return (scheme_make_char (SCHEME_INT_VAL (argv[0])));
}

static Scheme_Object *
char_upcase (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char-upcase");

  return (scheme_make_char (toupper (SCHEME_CHAR_VAL (argv[0]))));
}

static Scheme_Object *
char_downcase (int argc, Scheme_Object *argv[])
{
  CHAR_UN_CHECK("char-downcase");

  return (scheme_make_char (tolower (SCHEME_CHAR_VAL (argv[0]))));
}

