/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

#define GLOBAL_TABLE_SIZE 5313

/* globals */
int scheme_constant_builtins;
int scheme_allow_set_undefined;
int scheme_no_keywords;

int scheme_starting_up;

/* locals */
static Scheme_Env *scheme_make_env (void);
static Scheme_Object *constant_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *undefine(int argc, Scheme_Object *argv[]);
static Scheme_Object *list_globals(int argc, Scheme_Object *argv[]);

static Scheme_Object *write_variable(Scheme_Object *obj);
static Scheme_Object *read_variable(Scheme_Object *obj);
static Scheme_Object *write_stat_dist(Scheme_Object *obj);
static Scheme_Object *read_stat_dist(Scheme_Object *obj);
static Scheme_Object *read_stat_dist_unbox(Scheme_Object *obj);

static void scheme_hash_percent(const char *name, Scheme_Object *obj, 
				Scheme_Env *);

static int annotate_loc_id = 0;

Scheme_Env *scheme_basic_env ()
{
  scheme_starting_up = 1;

  scheme_init_stack_check();

  scheme_make_process();
  scheme_current_process->stack_start = NULL;
  scheme_top_level_env(scheme_current_process);

  scheme_starting_up = 0;

  return scheme_env;
}

void scheme_primitive_through_scheme(const char *name, Scheme_Env *env)
{
  if (scheme_constant_builtins)
    scheme_constant(scheme_intern_symbol(name), env);

  scheme_hash_percent(name, scheme_lookup_global(scheme_intern_symbol(name), 
						 env),
		      env);
}

void scheme_top_level_env(Scheme_Process *process)
{
  Scheme_Env *env;
  Scheme_Process *old;

  old = scheme_current_process;
  scheme_current_process = process;

  env = scheme_make_env ();
  process->globals = ((Scheme_Start_Env *)env)->globals;

  /* The ordering of the first few init calls is important. 
     Add to the end of the list, not the beginning. */
  scheme_init_symbol_table ();
  scheme_init_type_symbol_table ();
  scheme_init_type (env);
  scheme_init_symbol_type (env);
  scheme_init_fun (env);
  scheme_init_symbol (env);
  scheme_init_type_symbol (env);
  scheme_init_list (env);
  scheme_init_number (env);
  scheme_init_port (env);
  scheme_init_string (env);
  scheme_init_vector (env);
  scheme_init_char (env);
  scheme_init_bool (env);
  scheme_init_syntax (env);
  scheme_init_eval (env);
  scheme_init_error (env);
  scheme_init_promise (env);
  scheme_init_struct (env);
  scheme_init_debug(env);
  scheme_init_process(env);
  scheme_init_trigger(env);
  scheme_init_dynamic_extension(env);
  scheme_regexp_initialize(env);

  scheme_add_global_constant ("constant-name", 
			      scheme_make_prim_w_arity(constant_prim, 1, 1), 
			      env);
  scheme_add_global_constant ("undefine", 
			      scheme_make_prim_w_arity(undefine, 1, 1), 
			      env);
  scheme_add_global_constant ("make-global-value-list", 
			      scheme_make_prim_w_arity(list_globals, 0, 0), 
			      env);

  scheme_env = env;

  scheme_init_object(env);

  if (scheme_starting_up) {
    scheme_install_type_writer(scheme_variable_type, write_variable);
    scheme_install_type_reader(scheme_variable_type, read_variable);
    scheme_install_type_writer(scheme_static_distance_type, write_stat_dist);
    scheme_install_type_reader(scheme_static_distance_type, read_stat_dist);
    scheme_install_type_writer(scheme_static_distance_unbox_type, write_stat_dist);
    scheme_install_type_reader(scheme_static_distance_unbox_type, read_stat_dist_unbox);
  }

#define EVAL_ONE_STR(str) scheme_eval_string(str, env)
#define JUST_DEFINED(name) scheme_primitive_through_scheme(#name, env)

#include "macro.inc"

#ifdef TIME_SYNTAX
  scheme_eval_string("(defmacro time (expr) "
		     "`(#%time-apply (lambda () ,expr)))", env);
  scheme_primitive_through_scheme("time", env);
#endif

  scheme_current_process = old;
}

static Scheme_Object *
undefine(int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("undefine", 1, 1, argc, argv);
  if (!SCHEME_SYMBOLP(argv[0]))
    scheme_wrong_type("undefine", "symbol", 0, argc, argv);

  scheme_remove_global_symbol(argv[0], scheme_env);
  return scheme_void;
}

static Scheme_Env *
scheme_make_env (void)
{
  Scheme_Hash_Table *globals;
  Scheme_Start_Env *env;

  globals = scheme_hash_table (GLOBAL_TABLE_SIZE, SCHEME_hash_ptr, 
			       1, scheme_starting_up);

  env = (Scheme_Start_Env *)scheme_malloc(sizeof(Scheme_Start_Env));
  env->basic.num_bindings = 0;
  env->basic.next = NULL;
  env->globals = globals;

  return (Scheme_Env *)env;
}

static Scheme_Hash_Table *get_globals(Scheme_Env *env)
{
   while (env->basic.next)
    env = env->basic.next;
  return ((Scheme_Start_Env *)env)->globals;
}

static void
scheme_do_add_global_symbol(Scheme_Env *env, Scheme_Object *sym, 
			    Scheme_Object *obj, int constant)
{
  Scheme_Hash_Table *globals;

  globals = get_globals(env);

  scheme_add_to_table(globals, (char *)sym, obj, constant);
}

static void
scheme_do_add_global(Scheme_Env *env, const char *name, 
		     Scheme_Object *obj, int constant)
{
  Scheme_Object *sym;
  
  sym = scheme_intern_symbol(name);
  scheme_do_add_global_symbol(env, sym, obj, constant);
}

void
scheme_add_global(const char *name, Scheme_Object *obj, Scheme_Env *env)
{
  scheme_do_add_global(env, name, obj, 0);
}

void
scheme_add_global_symbol(Scheme_Object *sym, Scheme_Object *obj, 
			 Scheme_Env *env)
{
  scheme_do_add_global_symbol(env, sym, obj, 0);
}

void
scheme_remove_global_symbol(Scheme_Object *sym, Scheme_Env *env)
{
  Scheme_Bucket *b;
  Scheme_Hash_Table *globals;

  globals = get_globals(env);

  b = scheme_bucket_from_table(globals, (char *)sym);

  if (!b)
    return;

  if (((Scheme_Bucket_With_Const_Flag *)b)->const_val)
    scheme_signal_error("undefine: can't undefine constant %s",
			SCHEME_SYM_VAL(sym));

  b->val = NULL;
}

void
scheme_remove_global(const char *name, Scheme_Env *env)
{
  Scheme_Object *sym;

  sym = scheme_intern_symbol(name);

  scheme_remove_global_symbol(sym, env);
}

void
scheme_add_global_constant(const char *name, Scheme_Object *obj, 
			   Scheme_Env *env)
{
  scheme_do_add_global(env, name, obj, scheme_constant_builtins);
  scheme_hash_percent(name, obj, env);
}

void
scheme_add_global_keyword(const char *name, Scheme_Object *obj, 
			  Scheme_Env *env)
{
  scheme_do_add_global(env, name, obj, scheme_constant_builtins);

  if (!scheme_no_keywords) {
    Scheme_Object *sym;
    Scheme_Bucket *b;
    Scheme_Hash_Table *globals;

    globals = get_globals(env);

    sym = scheme_intern_symbol(name);
    b = scheme_bucket_from_table(globals, (char *)sym);
    ((Scheme_Bucket_With_Const_Flag *)b)->keyword = 1;
  }
}

static void scheme_hash_percent(const char *name, Scheme_Object *obj,
				Scheme_Env *env)
{
  int len;
  Scheme_Object *sym;
  Scheme_Bucket *b;
  static int buflen = 0;
  static char *buffer;

  len = strlen(name);
  if (len + 2 > buflen) {
    buflen = 2 * (buflen + len + 2);
    buffer = scheme_malloc_atomic(buflen + 1);
    buffer[0] = '#';
    buffer[1] = '%';
  }

  memcpy(buffer + 2, name, len + 1);
  scheme_do_add_global(env, buffer, obj, 1);

  if (!scheme_no_keywords) {
    Scheme_Hash_Table *globals;

    globals = get_globals(env);

    sym = scheme_intern_symbol(buffer);
    b = scheme_bucket_from_table(globals, (char *)sym);
    ((Scheme_Bucket_With_Const_Flag *)b)->keyword = 1;
  }
}

void scheme_check_identifier(const char *form, Scheme_Object *id, 
			     const char *where, Scheme_Env *env)
{
  Scheme_Bucket *b;
  char *s;
  Scheme_Hash_Table *globals;

  if (!where)
    where = "";

  if (!SCHEME_SYMBOLP(id))
    scheme_signal_error("%s: identifier must be a symbol%s", form, where);
  
  if (scheme_no_keywords)
    return;
  
  globals = get_globals(env);

  if (scheme_lookup_in_table(globals, (char *)id)) {
    b = scheme_bucket_from_table(globals, (char *)id);
    if (((Scheme_Bucket_With_Const_Flag *)b)->keyword)
      scheme_signal_error("%s: illegal use of keyword `%s'%s",
			  form, id, where);
  }

  s = SCHEME_SYM_VAL(id);
  if (s[0] == '#' && s[1] == '%')
    scheme_signal_error("%s: illegal use of special symbol `%s'%s",
			form, s, where);
}

Scheme_Env *
scheme_new_special_frame(int num_bindings, int flags)
{
  Scheme_Env *frame;
  int count;
  
  count = num_bindings;
  if (flags & SCHEME_ANNOTATING)
    count *= 2;

  frame = (Scheme_Env *)scheme_malloc(sizeof(Scheme_Basic_Env)
				      + (count * sizeof (Scheme_Object*)));
  frame->basic.num_bindings = num_bindings;
  frame->basic.autounbox = flags & SCHEME_AUTO_UNBOX;
  frame->basic.flags = ((flags & SCHEME_ANNOTATING) ? SCHEME_HAS_ANNOTATIONS : 0);

  return frame;
}

Scheme_Env *scheme_new_compilation_frame(int num_bindings, int flags)
{
  return scheme_new_special_frame(num_bindings, flags);
}

void
scheme_add_binding(int index, Scheme_Object *val, Scheme_Env *frame)
{
  if ((index >= frame->basic.num_bindings) || (index < 0))
    scheme_signal_error ("internal error: scheme_add_binding: "
			 "index out of range: %d", index);
  
  frame->values[index] = val;
}

void scheme_add_compilation_binding(int index, Scheme_Object *val, 
				    Scheme_Env *frame)
{
  scheme_add_binding(index, val, frame);
  if (frame->basic.flags & SCHEME_ANNOTATING) {
    annotate_loc_id++;
    frame->values[index + frame->basic.num_bindings] = 
      scheme_make_pair(val, scheme_make_integer(annotate_loc_id));
  }
}

Scheme_Env *
scheme_add_frame(Scheme_Object *vals, Scheme_Env *env, int flags)
{
  Scheme_Env *frame;
  int len, i, count;
  

  count = len = scheme_list_length(vals);
  if (flags & SCHEME_ANNOTATING)
    count *= 2;
  frame = (Scheme_Env *)scheme_malloc(sizeof(Scheme_Basic_Env)
				      + (count * sizeof (Scheme_Object*)));
  frame->basic.num_bindings = len;
  frame->basic.autounbox = flags & SCHEME_AUTO_UNBOX;
  frame->basic.flags = ((flags & SCHEME_ANNOTATING) ? SCHEME_HAS_ANNOTATIONS : 0);

  for (i = 0; i < len ; i++)
    if (SCHEME_SYMBOLP(vals))
      frame->values[i] = vals;
    else {
      frame->values[i] = SCHEME_CAR (vals);
      vals = SCHEME_CDR (vals);
    }
  
  if (flags & SCHEME_ANNOTATING)
    for (i = 0; i < len ; i++) {
      annotate_loc_id++;
      frame->values[i + len] = 
	scheme_make_pair(frame->values[i], scheme_make_integer(annotate_loc_id));
    }

  frame->basic.next = env;

  return frame;
}

Scheme_Env *scheme_add_compilation_frame(Scheme_Object *vals, 
					 Scheme_Env *env, int flags)
{
  return scheme_add_frame(vals, env, flags);
}

Scheme_Env *scheme_lengthen_compilation_frame(int count, Scheme_Env *env)
{
  int i, flags, total;
  Scheme_Env *frame;

  flags = (env->basic.flags & SCHEME_HAS_ANNOTATIONS) ? SCHEME_ANNOTATING : 0;

  total = env->basic.num_bindings + count;

  frame = scheme_new_compilation_frame(total, flags);

  for (i = env->basic.num_bindings; i--; )  {
    frame->values[i] = env->values[i];
    if (flags)
      frame->values[total] = env->values[env->basic.num_bindings + i];
  }

  frame = scheme_extend_env(frame, env->basic.next);

  return frame;
}

void scheme_frame_backinfo(Scheme_Env *orig, Scheme_Env *extended)
{
  if (orig && !(orig->basic.flags & SCHEME_HAS_SET_BANG))
    orig->basic.flags |= (extended->basic.flags & SCHEME_HAS_SET_BANG);
}

Scheme_Object *scheme_annotate_frame_id(Scheme_Env *frame, Scheme_Object *name)
{
  int i;

  for (i = frame->basic.num_bindings; i--; )
    if (SAME_OBJ(frame->values[i], name))
      return frame->values[i + frame->basic.num_bindings];

  return scheme_void;
}

void 
scheme_set_value (Scheme_Object *var, Scheme_Object *val, Scheme_Env *env)
{
  Scheme_Type type;
  Scheme_Bucket *b;
  Scheme_Env *frame;
  int i, j;

  type = SCHEME_TYPE(var);
  if (SAME_TYPE(type, scheme_variable_type)) {
    b = SCHEME_VAR_BUCKET(var);

    if (b->val) {
      if (((Scheme_Bucket_With_Const_Flag *)b)->const_val)
	scheme_signal_error("set!: cannot modifiy value of constant %s",
			    SCHEME_SYM_VAL(b->key));
      
      b->val = val;
    } else if (scheme_allow_set_undefined)
      b->val = val;
    else
      scheme_signal_error("can't set! unbound symbol: %s", 
			  SCHEME_SYM_VAL(b->key));
  } else if (SAME_TYPE(type, scheme_static_distance_type)
	     || SAME_TYPE(type, scheme_static_distance_unbox_type)) {
    frame = env;
    j = SCHEME_SDFRAME_VAL(var);
    i = SCHEME_SDPOS_VAL(var);
    while (j--)
      frame = frame->basic.next;
    if (SAME_TYPE(type, scheme_static_distance_unbox_type))
      SCHEME_PTR_VAL(frame->values[i]) = val;
    else
      frame->values[i] = val;
  } else
    scheme_signal_error("bad object for set!");
}

Scheme_Object *
scheme_static_distance (Scheme_Object *symbol, Scheme_Env *env,
			int elim_const_lookup)
{
  Scheme_Env *frame;
  Scheme_Object *v;
  int j = 0;
  Scheme_Bucket *b;
  Scheme_Hash_Table *globals;
  
  frame = env;
  while ( frame->basic.next != NULL )
    {
      int i;

      for ( i = frame->basic.num_bindings ; i--; )
	{
	  if (SAME_OBJ(symbol, frame->values[i]))
	    {
	      v = (Scheme_Object *)
		scheme_malloc_atomic(sizeof(Scheme_Static_Dist));
	      v->type = frame->basic.autounbox 
		? scheme_static_distance_unbox_type
		  : scheme_static_distance_type;
	      SCHEME_SDFRAME_VAL(v) = j;
	      SCHEME_SDPOS_VAL(v) = i;
	      return v;
	    }
	}
      frame = frame->basic.next;
      j++;
    }

  globals = get_globals(env);
  b = scheme_bucket_from_table(globals, (char *)symbol);
  if (elim_const_lookup && b && b->val 
      && ((Scheme_Bucket_With_Const_Flag *)b)->const_val)
    return b->val;

  return (Scheme_Object *)scheme_global_bucket(symbol, frame);
}

Scheme_Object *
scheme_lookup_value(Scheme_Object *distance, Scheme_Env *env)
{
  Scheme_Env *frame;
  Scheme_Object *v;
  Scheme_Type type;
  int j;
  int i;

  type = SCHEME_TYPE(distance);
  if (NOT_SAME_TYPE(type, scheme_static_distance_type)
      && NOT_SAME_TYPE(type, scheme_static_distance_unbox_type))
    return NULL;

  frame = env;
  j = SCHEME_SDFRAME_VAL(distance);
  i = SCHEME_SDPOS_VAL(distance);
  while (frame->basic.next != NULL )
    {
      if (!j) {
	if (i >= frame->basic.num_bindings)
	  scheme_signal_error("internal error: bad static distance lookup "
			      "(index)");

	v = frame->values[i];
	if (SAME_TYPE(type, scheme_static_distance_unbox_type))
	  v = (Scheme_Object *)SCHEME_PTR_VAL(v);
	return v;
      }
      frame = frame->basic.next;
      --j;
    }

  scheme_signal_error("internal error: bad static distance lookup (frame)");

  return scheme_void;
}

Scheme_Object *
scheme_lookup_global (Scheme_Object *symbol, Scheme_Env *env)
{
  return scheme_lookup_in_table(get_globals(env), (char *)symbol);
}

Scheme_Bucket *
scheme_global_bucket (Scheme_Object *symbol, Scheme_Env *env)
{
  return (scheme_bucket_from_table(get_globals(env), (char *)symbol));
}

Scheme_Object *scheme_make_envunbox(Scheme_Object *value)
{
  Scheme_Object *obj;

  obj = scheme_alloc_small_object();
  obj->type = scheme_envunbox_type;
  SCHEME_PTR_VAL(obj) = value;

  return obj;
}

void scheme_constant(Scheme_Object *sym, Scheme_Env *env)
{
  if (!SCHEME_SYMBOLP(sym))
    scheme_wrong_type("constant-name", "symbol", 0, 1, &sym);

  scheme_do_add_global_symbol(env, sym, NULL, 1);
}

static Scheme_Object *
constant_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("constant-name", 1, 1, argc, argv);

  scheme_constant(argv[0], scheme_env);
  return scheme_void;
}

static Scheme_Object *list_globals(int argc, Scheme_Object *argv[])
{
  Scheme_Object *l = scheme_null;
  int i;
  Scheme_Hash_Table *globals;
  Scheme_Bucket *b;

  if (argc)
    scheme_wrong_count("make-global-value-list", 0, 0, argc, argv);

  globals = scheme_current_process->globals;

  for (i = 0; i < globals->size; i++) {
    b = globals->buckets[i];
    if (b && b->val)
      l = scheme_make_pair(scheme_make_pair((Scheme_Object *)b->key, b->val),
			   l);
  }

  return l;
}

/*********************************************************************/

static Scheme_Object *write_variable(Scheme_Object *obj)
{
  Scheme_Object *sym;

  sym = (Scheme_Object *)(SCHEME_VAR_BUCKET(obj))->key;
  return scheme_make_pair(sym, scheme_null);
}

static Scheme_Object *read_variable(Scheme_Object *obj)
{
  if (!SCHEME_PAIRP(obj) || !SCHEME_NULLP(SCHEME_CDR(obj))
      || !SCHEME_SYMBOLP(SCHEME_CAR(obj)))
    scheme_signal_error("bad compiled global");

  return scheme_static_distance(SCHEME_CAR(obj), scheme_env, 1);
}

static Scheme_Object *write_stat_dist(Scheme_Object *obj)
{
  return scheme_make_pair(scheme_make_integer(SCHEME_SDFRAME_VAL(obj)),
			  scheme_make_integer(SCHEME_SDPOS_VAL(obj)));
}

static Scheme_Object *read_stat_dist_gen(Scheme_Object *obj, Scheme_Type type)
{
  Scheme_Object *v;

  if (!SCHEME_PAIRP(obj) || !SCHEME_INTP(SCHEME_CDR(obj))
      || !SCHEME_INTP(SCHEME_CAR(obj)))
    scheme_signal_error("bad compiled local");

  v = (Scheme_Object *)scheme_malloc_atomic(sizeof(Scheme_Static_Dist));
  v->type = type;
  SCHEME_SDFRAME_VAL(v) = SCHEME_INT_VAL(SCHEME_CAR(obj));
  SCHEME_SDPOS_VAL(v) = SCHEME_INT_VAL(SCHEME_CDR(obj));

  return v;
}

static Scheme_Object *read_stat_dist(Scheme_Object *obj)
{
  return read_stat_dist_gen(obj, scheme_static_distance_type);
}

static Scheme_Object *read_stat_dist_unbox(Scheme_Object *obj)
{
  return read_stat_dist_gen(obj, scheme_static_distance_unbox_type);
}

