/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <stdlib.h>
#include <ctype.h>

#define MAX_NUMBER_SIZE 500
#define MAX_SYMBOL_SIZE 255

short scheme_can_read_compiled;

/* local function prototypes */

static Scheme_Object *read_list (Scheme_Object *port, char closer);
static Scheme_Object *read_string (Scheme_Object *port);
static Scheme_Object *read_quote (Scheme_Object *port);
static Scheme_Object *read_vector (Scheme_Object *port, char closer, long len);
static Scheme_Object *read_type_symbol (Scheme_Object *port);
static Scheme_Object *read_number (Scheme_Object *port, int, int, int, int);
static Scheme_Object *read_symbol (Scheme_Object *port, int type_symbol);
static Scheme_Object *read_character (Scheme_Object *port);
static Scheme_Object *read_quasiquote (Scheme_Object *port);
static Scheme_Object *read_unquote (Scheme_Object *port);
static Scheme_Object *read_unquote_splicing (Scheme_Object *port);
static Scheme_Object *read_box (Scheme_Object *port);
static Scheme_Object *read_compiled (Scheme_Object *port);
static int skip_whitespace_comments (Scheme_Object *port);
static int peek_char (Scheme_Object *port);
static int match_chars (Scheme_Object *port, char *str);

Scheme_Object *
scheme_read (Scheme_Object *port)
{
  int ch, ch2, depth;
  long vector_length;

 start_over:
  ch = scheme_getc (port);
  while (isspace(ch))
    ch = scheme_getc (port);

  switch ( ch )
    {
    case EOF: return (scheme_eof);
    case ')': scheme_signal_error("read: unexpected ')' at %ld, "
				  "line %ld in %s", 
				  scheme_tell(port),
				  scheme_tell_line(port), 
				  SCHEME_IPORT_NAME(port));
    case ']': scheme_signal_error("read: unexpected ']' at %ld, "
				  "line %ld in %s", 
				  scheme_tell(port),
				  scheme_tell_line(port), 
				  SCHEME_IPORT_NAME(port));
    case '(': return (read_list (port, ')'));
    case '[': return (read_list (port, ']'));
    case '"': return (read_string (port));
    case '\'': return (read_quote (port));
    case '`': return (read_quasiquote (port));
    case ',':
      if (peek_char (port) == '@') {
	ch = scheme_getc (port);
	return (read_unquote_splicing (port));
      } else
	return (read_unquote (port));
    case ';':
      while (((ch = scheme_getc (port)) != '\n') && (ch != '\r')) {
	if (ch == EOF)
	  return (scheme_eof);
      }
      goto start_over;
    case '+':
    case '-':
    case '.':
      ch2 = peek_char (port);
      if (isdigit(ch2) || (ch2 == '.')) {
	scheme_ungetc (ch, port);
	return read_number(port, 0, 0, 10, 0);
      } else {
	scheme_ungetc (ch, port);
	return (read_symbol (port, 0));
      }
    case '#':
      ch = scheme_getc (port);
      switch ( ch )
	{
	case '%':
	  scheme_ungetc('%', port);
	  scheme_ungetc('#', port);
	  return read_symbol(port, 0);
	case '<': return read_type_symbol(port);
	case '(': return read_vector(port, ')', -1);
	case '[': return read_vector(port, ']', -1);
	case '\\': return read_character(port);
	case 't': return scheme_true;
	case 'f': return scheme_false;
	case 'x': return read_number(port, 0, 0, 16, 1);
	case 'b': return read_number(port, 0, 0, 2, 1);
	case 'o': return read_number(port, 0, 0, 8, 1);
	case 'e': return read_number(port, 0, 1, 10, 0);
	case 'i': return read_number(port, 1, 0, 10, 0);
	case '`': 
	  if (scheme_can_read_compiled)
	    return read_compiled(port);
	  else
	    scheme_signal_error ("read: #` expressions not currently enabled");
	case '|':
	  depth = 0;
	  ch2 = 0;
	  do {
	    ch = scheme_getc (port);
	    if (ch == EOF)
	      scheme_signal_error ("read: end of file in #| comment");
	    if ((ch2 == '|') && (ch == '#')) {
	      if (!(depth--))
		goto start_over;
	    } else if ((ch2 == '#') && (ch == '|'))
	      depth++;
	    ch2 = ch;
	  } while (1);
	  break;
	case '&':
	  return read_box(port);
	default:
	  vector_length = -1;
	  while (isdigit(ch)) {
	    if (vector_length < 0)
	      vector_length = 0;
	    vector_length = (vector_length * 10) + (ch - 48);
	    ch = scheme_getc(port);
	  }
	  if (ch == '(')
	    return (read_vector (port, ')', vector_length));
	  if (ch == '[')
	    return (read_vector (port, ']', vector_length));
	   
	  scheme_signal_error ("read: bad syntax `#%c' at %ld, line %ld in %s",
			       ch,
			       scheme_tell(port),
			       scheme_tell_line(port),
			       SCHEME_IPORT_NAME(port));
	}
    default:
      scheme_ungetc (ch, port);
      if (isdigit (ch))
	return (read_number(port, 0, 0, 10, 0));
      else
	return (read_symbol (port, 0));
    }
}

/* "(" has already been read */
static Scheme_Object *
read_list (Scheme_Object *port, char closer)
{
  Scheme_Object *list = NULL, *last = NULL, *car, *cdr;
  int ch;
  long start, startline;

  start = scheme_tell(port);
  startline =scheme_tell_line(port);

  while (1) {
    ch = skip_whitespace_comments (port);
    if (ch == EOF)
      scheme_signal_error("read: expected a '%c'; started at "
			  "position %ld, line %ld in %s", closer, start, 
			  startline, SCHEME_IPORT_NAME(port));

    if (ch == closer)
      return list ? list : scheme_null;

    scheme_ungetc(ch, port);
    car = scheme_read(port);

    ch = skip_whitespace_comments(port);
    if (ch == closer) {
      cdr = scheme_make_pair(car, scheme_null);
      if (!list)
	list = cdr;
      else
	SCHEME_CDR(last) = cdr;
      return list;
    } else if ((ch == '.') && isspace(peek_char(port))) {
      cdr = scheme_read (port);
      ch = skip_whitespace_comments (port);
      if (ch != closer)
	scheme_signal_error ("read: malformed list");
      cdr = scheme_make_pair(car, cdr);
      if (!list)
	list = cdr;
      else
	SCHEME_CDR(last) = cdr;
      return list;
    } else {
      scheme_ungetc(ch, port);
      cdr = scheme_make_pair(car, scheme_null);
      if (!list)
	list = cdr;
      else
	SCHEME_CDR(last) = cdr;
      last = cdr;
    }
  }
}

/* '"' has already been read */
static Scheme_Object *
read_string (Scheme_Object *port)
{
  char *buf, *oldbuf, onstack[32];
  int i, ch;
  long start, startline;
  long size = 31, oldsize;

  i = 0;
  start = scheme_tell(port);
  startline = scheme_tell_line(port);
  buf = onstack;
  while ((ch = scheme_getc (port)) != '"') {
    if (ch == EOF)
      scheme_signal_error("read: expected a '\"'; started at position %ld"
			  ", line %ld in %s", 
			  start, startline, SCHEME_IPORT_NAME(port));
    if (ch == '\\')
      ch = scheme_getc (port);

    if (i >= size)
      { 
	oldsize = size;
	oldbuf = buf;

	size *= 2;
	buf = (char *)scheme_malloc_atomic(size + 1);
	memcpy(buf, oldbuf, oldsize);
      }
    buf[i++] = ch;
  }
  buf[i] = '\0';

  if (i <= 31)
    return scheme_make_string(buf);
  else
    return scheme_make_string_without_copying(buf);
}

/* "'" has been read */
static Scheme_Object *
read_quote (Scheme_Object *port)
{
  Scheme_Object *obj;

  obj = scheme_read (port);
  return (scheme_make_pair (scheme_quote_symbol, 
			    scheme_make_pair (obj, scheme_null)));
}

/* "#(" has been read */
static Scheme_Object *
read_vector (Scheme_Object *port, char closer, long requestLength)
{
  Scheme_Object *obj, *vec;
  int len, i;
  
  obj = read_list(port, closer);
  len = scheme_list_length (obj);
  if (requestLength >= 0 && len > requestLength)
    scheme_signal_error("specified vector length is too small");
  if (requestLength < 0)
    requestLength = len;
  vec = scheme_make_vector (requestLength, NULL);
  for ( i=0 ; i<len ; ++i )
    {
      SCHEME_VEC_ELS(vec)[i] = SCHEME_CAR(obj);
      obj = SCHEME_CDR(obj);
    }
  if (i < requestLength) {
    if (len)
      obj = SCHEME_VEC_ELS(vec)[len - 1];
    else
      obj = scheme_make_integer(0);
    for (; i < requestLength; i++)
      SCHEME_VEC_ELS(vec)[i] = obj;
  }
  return (vec);
}

/* nothing has been read, except maybe some flags */
static Scheme_Object  *
read_number (Scheme_Object *port, int is_float, int is_not_float, 
	     int radix, int radix_set)
{
  char *buf, *oldbuf, onstack[32];
  int i, ch;
  long size = 31, oldsize;
  Scheme_Object *o;

  i = 0;
  buf = (char *)onstack;
  while ((ch = scheme_getc (port)) != EOF) {
    if (isspace(ch)
	|| (ch == '(')
	|| (ch == '[')
	|| (ch == ')')
	|| (ch == ']')
	|| (ch == '"')
	|| (ch == ';')) {
      scheme_ungetc(ch, port);
      break;
    }

    if (i >= size) {
      oldsize = size;
      oldbuf = buf;

      size *= 2;
      buf = (char *)scheme_malloc_atomic(size + 1);
      memcpy(buf, oldbuf, oldsize);
    }
    buf[i++] = ch;
  }
  buf[i] = '\0';

  o = scheme_read_number(buf, is_float, is_not_float, radix, radix_set, 1);
  if (SAME_OBJ(o, scheme_false))
    return scheme_intern_symbol(buf);
  else
    return o;
}

/* nothing has been read */
static Scheme_Object *
read_symbol (Scheme_Object *port, int type_symbol)
{
  char *buf, *oldbuf, onstack[MAX_SYMBOL_SIZE];
  int size, oldsize;
  int i, ch, nvdeep = 0;

  i = 0;
  size = MAX_SYMBOL_SIZE - 1;
  buf = onstack;

  while (((ch = scheme_getc (port)) != '(')
	 && (!isspace(ch) || (type_symbol && (ch == ' ')))
	 && (ch != '[')
	 && (ch != ')')
	 && (ch != ']')
	 && (ch != '"')
	 && (ch != ';')
	 && (!type_symbol || ch != '>' || nvdeep)
	 && (ch != EOF)) {
    if (i >= size) {
      oldsize = size;
      oldbuf = buf;
      
      size *= 2;
      buf = (char *)scheme_malloc_atomic(size + 1);
      memcpy(buf, oldbuf, oldsize);
    }

    buf[i++] = ch;

    if (type_symbol) {
      if (ch == '>')
	--nvdeep;
      else if (ch == '<')
	nvdeep++;
    }
  }

  if (ch != EOF)
    scheme_ungetc (ch, port);

  buf[i] = '\0';

  return scheme_intern_symbol((char *)buf);
}

/* '#<' has been read */
static Scheme_Object *read_type_symbol(Scheme_Object *port)
{
  Scheme_Object *sym;
  long p, l;

  p = scheme_tell(port);
  l = scheme_tell_line(port);

  sym = read_symbol(port, 1);
  if (scheme_getc(port) != '>')
    scheme_signal_error("read: expected a `>'; started "
			"at %ld, line %ld in %s",
			p, l, SCHEME_IPORT_NAME(port));

  return scheme_intern_type_symbol(sym);
}

/* "#\" has been read */
static Scheme_Object *
read_character (Scheme_Object *port)
{
  int ch, next;

  ch = scheme_getc (port);
  next = peek_char(port);
  switch (ch)
    {
    case 'n': /* maybe `newline' or 'null' */
    case 'N':
      if ((next == 'e') || (next == 'E')) {
	if (! match_chars (port, "ewline"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return scheme_make_char('\n');
      } else if ((next == 'u') || (next == 'U')) {
	if (! match_chars (port, "ull"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return (scheme_make_char ('\0'));
      } else 
	return (scheme_make_char (ch));
    case 's': /* maybe `space' */
    case 'S':
      if ((next == 'p') || (next == 'P')) {
	if (! match_chars (port, "pace"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return (scheme_make_char (' '));
      } else
	return (scheme_make_char (ch));
    case 'r': /* maybe `rubout' */
    case 'R':
      if ((next == 'u') || (next == 'U')) {
	if (! match_chars (port, "ubout"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return (scheme_make_char (0x7f));
      } else if ((next == 'e') || (next == 'E')) {
	if (! match_chars (port, "eturn"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return (scheme_make_char ('\r'));
      } else
	return (scheme_make_char (ch));
    case 'p': /* maybe `page' */
    case 'P':
      if ((next == 'a') || (next == 'A')) {
	if (! match_chars (port, "age"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return (scheme_make_char ('\f'));
      } else
	return (scheme_make_char (ch));
    case 't': /* maybe `tab' */
    case 'T':
      if ((next == 'a') || (next == 'A')) {
	if (! match_chars (port, "ab"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return (scheme_make_char ('\t'));
      } else
	return (scheme_make_char (ch));
    case 'b': /* maybe `backspace' */
    case 'B':
      if ((next == 'a') || (next == 'A')) {
	if (! match_chars (port, "ackspace"))
	  scheme_signal_error ("read: bad character constant", NULL);
	else
	  return (scheme_make_char ('\b'));
      } else
	return (scheme_make_char (ch));
    case 'l': /* maybe `linefeed' */
    case 'L':
      if ((next == 'i') || (next == 'I')) {
	if (! match_chars (port, "inefeed"))
	  scheme_signal_error ("read: bad character constant");
	else
	  return scheme_make_char('\n');
      } else
	return (scheme_make_char (ch));
    default:
      return (scheme_make_char (ch));
    }
}

/* "`" has been read */
static Scheme_Object *
read_quasiquote (Scheme_Object *port)
{
  Scheme_Object *quoted_obj, *ret;
  
  quoted_obj = scheme_read (port);
  ret = scheme_make_pair (scheme_quasiquote_symbol, 
			  scheme_make_pair (quoted_obj, scheme_null));
  return (ret);
}

/* "," has been read */
static Scheme_Object *
read_unquote (Scheme_Object *port)
{
  Scheme_Object *obj, *ret;

  obj = scheme_read (port);
  ret = scheme_make_pair (scheme_unquote_symbol, 
			  scheme_make_pair (obj, scheme_null));
  return (ret);
}

/* ",@" has been read */
static Scheme_Object *
read_unquote_splicing (Scheme_Object *port)
{
  Scheme_Object *obj, *ret;

  obj = scheme_read (port);
  ret = scheme_make_pair (scheme_unquote_splicing_symbol, 
			  scheme_make_pair (obj, scheme_null));
  return (ret);
}

/* "#&" has been read */
static Scheme_Object *read_box(Scheme_Object *port)
{
  Scheme_Object *o;

  o = scheme_read(port);
  
  return scheme_box(o);
}

/* "#`" has been read */
static Scheme_Object *read_compiled (Scheme_Object *port)
{
  int ch = '0', type = 0;
  Scheme_Object *l;

  ch = scheme_getc(port);
  if (isdigit(ch)) {
    do {
      type = (type * 10) + (ch - '0');
      ch = scheme_getc(port);
    } while (isdigit(ch));

    if ((type < 0) || (type >= _scheme_last_type_))
      scheme_signal_error("bad #` type number");
  } else if (ch == '<') {
    Scheme_Object *ts;
    
    ts = read_type_symbol(port);
    type = scheme_find_type(ts);

    ch = scheme_getc(port);
  } else
    scheme_signal_error("bad #` expression");

  if (ch != '(' && ch != '[')
    scheme_signal_error("bad #` expression");

  l = read_list(port, (ch == '(') ? ')' : ']');

  if (!scheme_type_readers[type])
      scheme_signal_error("bad #` type number");

  return scheme_type_readers[type](l);
}

/* utilities */

static int
skip_whitespace_comments (Scheme_Object *port)
{
  int ch;

 start_over:

  while (isspace(ch = scheme_getc (port)));

  if (ch == ';') {
    do
      ch = scheme_getc (port);
    while (ch != '\n' && ch != '\r' && ch != EOF);
    goto start_over;
  }

  return ch;
}

static int
peek_char (Scheme_Object *port)
{
  int ch;

  ch = scheme_getc (port);
  scheme_ungetc (ch, port);
  return (ch);
}

static int
match_chars (Scheme_Object *port, char *str)
{
  int i;
  int ch;

  i = 0;
  while (str[i]) {
    ch = scheme_getc(port);
    if (tolower(ch) != tolower(str[i]))
      return (0);
    i++;
  }

  return 1;
}
