/*
 * File:        wx_cgrec.cc
 * Purpose:     wxChangeRecord implementations
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include <stdio.h>
#include "wx_media.h"
#include "wx_ptreq.h"

wxChangeRecord::wxChangeRecord(void)
{
}

wxChangeRecord::~wxChangeRecord(void)
{
}

Bool wxChangeRecord::Undo(wxMediaBuffer *)
{
  return FALSE;
}

Bool wxUnmodifyRecord::Undo(wxMediaBuffer *media)
{
  media->SetModified(FALSE);
  return FALSE;
}

wxInsertRecord::wxInsertRecord(long position, long length, Bool cont)
{
  start = position;
  end = position + length;
  continued = cont;
}

Bool wxInsertRecord::Undo(wxMediaBuffer *buffer)
{
  wxMediaEdit *media;
  media = (wxMediaEdit *)buffer;

  media->Delete(start, end);
  if (!continued)
   media->SetPosition(start);

  return continued;
}

wxInsertSnipRecord::wxInsertSnipRecord(wxSnip *thesnip, Bool cont)
{
  snip = thesnip;
  continued = cont;
}

Bool wxInsertSnipRecord::Undo(wxMediaBuffer *buffer)
{
  wxMediaPasteboard *media;
  media = (wxMediaPasteboard *)buffer;

  media->Delete(snip);
  if (!continued)
   media->SetSelected(snip);

  return continued;
}

class DeleteSnipItem /* : public wxObject --- uncomment to GC */
{
 public:
  ~DeleteSnipItem();

  wxDeleteSnipRecord *parent;
  wxSnip *snip, *before;
  float x, y;
};

DeleteSnipItem::~DeleteSnipItem()
{
  if (!parent->undid) {
#if !WXGARBAGE_COLLECTION_ON
    delete snip;
#else
    if (snip->flags & wxSNIP_OWNED)
      snip->flags -= wxSNIP_OWNED;
    snip->SetAdmin(NULL);
#endif
  }
}

wxDeleteSnipRecord::wxDeleteSnipRecord(Bool cont)
{
  continued = cont;
  deletions = new wxList();
}

wxDeleteSnipRecord::~wxDeleteSnipRecord()
{
  wxNode *node;

  for (node = deletions->First(); node; node = node->Next())
    delete (DeleteSnipItem *)node->Data();

  delete deletions;
}

void wxDeleteSnipRecord::InsertSnip(wxSnip *snip, wxSnip *before, 
				    float x, float y)
{
  DeleteSnipItem *item;

  item = new DeleteSnipItem();
  item->parent = this;
  item->snip = snip;
  item->before = before;
  item->x = x;
  item->y = y;

  deletions->Append((wxObject *)item);
}

Bool wxDeleteSnipRecord::Undo(wxMediaBuffer *buffer)
{
  wxNode *node;
  DeleteSnipItem *item;
  wxMediaPasteboard *media;

  media = (wxMediaPasteboard *)buffer;
  
  if (!continued)
    media->NoSelected();

  for (node = deletions->First(); node; node = node->Next()) {
    item = (DeleteSnipItem *)node->Data();

    /* Have to turn off the owned flag; we know that it's really ours */
    if (item->snip->flags & wxSNIP_OWNED)
      item->snip->flags -= wxSNIP_OWNED;

    media->Insert(item->snip, item->before, item->x, item->y);
    if (!continued)
      media->AddSelected(item->snip);
  }

  undid = TRUE;

  return continued;
}

wxDeleteRecord::wxDeleteRecord(long startpos, long endpos, Bool cont)
{
  continued = cont;
  start = startpos;
  end = endpos;
  undid = FALSE;
  deletions = new wxList();
  clickbacks = new wxList();
}

wxDeleteRecord::~wxDeleteRecord()
{
  if (!undid) {
#if !WXGARBAGE_COLLECTION_ON
    deletions->DeleteContents(TRUE);
#else
    wxSnip *snip;
    wxNode *node;
    for (node = deletions->First(); node; node = node->Next()) {
      snip = (wxSnip *)node->Data();
      if (snip->flags & wxSNIP_OWNED)
	snip->flags -= wxSNIP_OWNED;
      snip->SetAdmin(NULL);
    }
#endif
    clickbacks->DeleteContents(TRUE);
  }

  delete deletions;
  delete clickbacks;
}

void wxDeleteRecord::InsertSnip(wxSnip *snip)
{
  deletions->Append(snip);
}

void wxDeleteRecord::AddClickback(wxClickback *click)
{
  clickbacks->Append((wxObject *)click);
}

Bool wxDeleteRecord::Undo(wxMediaBuffer *buffer)
{
  wxNode *node;
  wxSnip *snip;
  wxMediaEdit *media;

  media = (wxMediaEdit *)buffer;
  
  for (node = deletions->First(); node; node = node->Next()) {
    snip = (wxSnip *)node->Data();

    /* Have to turn off the owned flag; we know that it's really ours */
    if (snip->flags & wxSNIP_OWNED)
      snip->flags -= wxSNIP_OWNED;

    media->Insert(snip, start);
  }
  for (node = clickbacks->First(); node; node = node->Next())
    media->SetClickback((wxClickback *)node->Data());

  if (!continued)
   media->SetPosition(end);

  undid = TRUE;

  return continued;
}

class StyleChange /* : public wxObject  */
{
 public:
  long start, end;
  wxStyle *style;
};

wxStyleChangeRecord::wxStyleChangeRecord(long startpos, long endpos, Bool cont)
{
  continued = cont;
  start = startpos;
  end = endpos;

  changes = new wxList();
}

wxStyleChangeRecord::~wxStyleChangeRecord()
{
  wxNode *node;

  for (node = changes->First(); node; node = node->Next())
    delete (StyleChange *)node->Data();

  delete changes;
}

void wxStyleChangeRecord::AddStyleChange(long start, long end, wxStyle *style)
{
  StyleChange *change;

  change = new StyleChange;

  change->start = start;
  change->end = end;
  change->style = style;

  changes->Append((wxObject *)change);
}

Bool wxStyleChangeRecord::Undo(wxMediaBuffer *buffer)
{
  wxNode *node;
  StyleChange *change;
  long p;
  wxMediaEdit *media;

  media = (wxMediaEdit *)buffer;
  
  p = start;

  for (node = changes->First(); node; node = node->Next()) {
    change = (StyleChange *)node->Data();
    media->ChangeStyle(change->style, change->start, change->end);
  }

  if (!continued)
    media->SetPosition(start, end);

  return continued;
}

class StyleChangeSnip : public wxObject 
{
 public:
  wxSnip *snip;
  wxStyle *style;
};

wxStyleChangeSnipRecord::wxStyleChangeSnipRecord(Bool cont)
{
  continued = cont;

  changes = new wxList();
}

wxStyleChangeSnipRecord::~wxStyleChangeSnipRecord()
{
  wxNode *node;
  
  for (node = changes->First(); node; node = node->Next())
    delete (StyleChange *)node->Data();
  
  delete changes;
}

void wxStyleChangeSnipRecord::AddStyleChange(wxSnip *snip, wxStyle *style)
{
  StyleChangeSnip *change;

  change = new StyleChangeSnip;

  change->snip = snip;
  change->style = style;

  changes->Append((wxObject *)change);
}

Bool wxStyleChangeSnipRecord::Undo(wxMediaBuffer *buffer)
{
  wxNode *node;
  StyleChangeSnip *change;
  wxMediaPasteboard *media;

  media = (wxMediaPasteboard *)buffer;
  
  if (!continued)
    media->NoSelected();

  for (node = changes->First(); node; node = node->Next()) {
    change = (StyleChangeSnip *)node->Data();
    media->ChangeStyle(change->style, change->snip);
    if (!continued)
      media->AddSelected(change->snip);
  }

  return continued;
}

wxMoveSnipRecord::wxMoveSnipRecord(wxSnip *s, float fx, float fy, 
				   Bool d, Bool cont)
{
  snip = s;
  continued = cont;
  x = fx;
  y = fy;
  delta = d;
}

Bool wxMoveSnipRecord::Undo(wxMediaBuffer *buffer)
{
  wxMediaPasteboard *media;

  media = (wxMediaPasteboard *)buffer;
  
  if (delta)
    media->Move(snip, x, y);
  else
    media->MoveTo(snip, x, y);

  return continued;
}

wxResizeSnipRecord::wxResizeSnipRecord(wxSnip *s, float fx, float fy, 
				       Bool cont)
{
  snip = s;
  continued = cont;
  x = fx;
  y = fy;
}

Bool wxResizeSnipRecord::Undo(wxMediaBuffer *buffer)
{
  wxMediaPasteboard *media;

  media = (wxMediaPasteboard *)buffer;

  media->Resize(snip, x, y);

  return continued;
}

