/*
 * File:        wx_media.cc
 * Purpose:     wxMediaEdit implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_media.h"
#include "wx_timer.h"

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "wx_mpriv.h"

static void StandardWordbreak(wxMediaEdit *win, long *start, long *end, int, void*);

wxMediaWordbreakMap wxTheMediaWordbreakMap;

/************************************************************************/

void wxInitMedia(void)
{
  wxInitClipboard();
  wxInitStyles();
}

wxMediaEdit::wxMediaEdit(float spacing, float *tabstops, int numtabs) 
{
  __type = wxTYPE_MEDIA_EDIT;
  bufferType = wxEDIT_BUFFER;

  locked = FALSE;
  userLocked = modified = FALSE;
  
  lineSpacing = spacing;
  maxWidth = minWidth = minHeight = maxHeight = -1;

  hiliteOn = TRUE;

  len = 0;

  AddEditorFunctions(map);

  startpos = endpos = 0;
  posateol = FALSE;

  flash = FALSE;
  flashTimer = NULL;

  delayRefresh = 0;

  refreshUnset = refreshBoxUnset = TRUE;
  refreshAll = FALSE;
  delayedscroll = -1;
  delayedscrollbox = FALSE;

  snipAdmin = new wxStandardSnipAdmin(this);

  caretStyle = NULL;

  MakeOnlySnip();

  snipCacheInvalid = FALSE;

  graphicsInvalid = flowInvalid = FALSE;
  graphicMaybeInvalid = TRUE;

  totalHeight = totalWidth = finalDescent = initialSpace = 0;

  bitmap = NULL;
  offscreen = new wxMemoryDC();
  bmHeight = bmWidth = 0;
  drawCachedInBitmap = FALSE;

  clickbacks = new wxList();

  typingStreak = FALSE;
  deletionStreak = FALSE;
  delayedStreak = FALSE;
  anchorStreak = FALSE;
  vcursorStreak = FALSE;
  killStreak = FALSE;

  changed = FALSE;

  tracking = dragging = FALSE;

  fileFormat = wxMEDIA_FF_STD;

  tabcount = numtabs;
  tabs = tabstops;
  tabSpace = wxTAB_WIDTH;
  tabSpaceInUnits = TRUE;

  wordBreak = StandardWordbreak;
  wordBreakData = NULL;
  wordBreakMap = &wxTheMediaWordbreakMap;
  if (!wxTheMediaWordbreakMap.IsUsed())
    wxTheMediaWordbreakMap.AdjustUsage(TRUE); /* Once to keep forever */
  wxTheMediaWordbreakMap.AdjustUsage(TRUE);

  caretLocationX = -1;
  caretOn = FALSE;
}

wxMediaEdit::~wxMediaEdit()
{
  wxSnip *snip, *next;

  SetWordbreakMap(NULL);

  for (snip = snips; snip; snip = next) {
    next = snip->next;
    delete snip;
  }

  delete lineRoot;

  clickbacks->DeleteContents(TRUE);
  delete clickbacks;

  if (tabs)
    delete[] tabs;

  offscreen->SelectObject(NULL);
  delete offscreen;
  if (bitmap)
    delete bitmap;

  delete snipAdmin;
}

wxMediaBuffer *wxMediaEdit::CopySelf(void)
{
  wxMediaEdit *m;

  m = new wxMediaEdit(lineSpacing);

  /* Copy parameters, such as tab settings: */
  if (tabs) {
    float *t;

    t = new float[tabcount];
    memcpy(t, tabs, sizeof(float) * tabcount);
    m->SetTabs(t, tabcount, tabSpace, tabSpaceInUnits);
  }

  /* Copy all the snips: */
  wxList *saveBuffer = wxmb_commonCopyBuffer, *copySnips;
  wxList *saveBuffer2 = wxmb_commonCopyBuffer2, *copySnips2;
  long savestart = startpos, saveend = endpos;

  wxmb_commonCopyBuffer = copySnips = new wxList();
  wxmb_commonCopyBuffer2 = copySnips2 = new wxList();
  startpos = 0; endpos = len;
  Copy();
  startpos = savestart; endpos = saveend;
  wxmb_commonCopyBuffer = saveBuffer;
  wxmb_commonCopyBuffer2 = saveBuffer2;

  wxNode *node;
  BeginEditSequence();

  for (node = copySnips->First(); node; node = node->Next())
    m->Insert((wxSnip *)node->Data());

  EndEditSequence();

  delete copySnips;
  delete copySnips2;

  m->SizeCacheInvalid();

  /* All done! */
  return m;
}

/******************************************************************/

void wxMediaEdit::OnEvent(wxMouseEvent &event)
{
  float scrollx, scrolly;
  float x, y;
  wxSnip *snip;
  wxDC *dc;
  Bool onit;
  Bool sequenced = FALSE;
  long now;

  if (!admin)
    return;

  if (!event.Moving())
    EndStreaks(wxSTREAK_KEY_SEQUENCE);

  if (event.ButtonDown() || caretSnip) {
    /* First, find clicked-on snip: */
    x = event.x;
    y = event.y;
    
    dc = admin->GetDC(&scrollx, &scrolly);
    y += scrolly;
    x += scrollx;

    if (!dc)
      return;
  } else
    dc = NULL;

  if (event.ButtonDown()) {
    now = FindPosition(x, y, NULL, &onit);
    if (onit)
      snip = FindSnip(now, +1);
    else
      snip = NULL;
    sequenced = (PTRNE(snip, caretSnip));
    if (sequenced)
      BeginEditSequence();
    SetCaretOwner(snip);
  }

  if (caretSnip && (caretSnip->flags & wxSNIP_HANDLES_EVENTS)) {
    GetSnipPositionAndLocation(caretSnip, NULL, &x, &y);
    caretSnip->OnEvent(dc, x - scrollx, y - scrolly, x, y, event);
    if (sequenced)
      EndEditSequence();
    return;
  }

  OnLocalEvent(event);

  if (sequenced)
    EndEditSequence();
}

void wxMediaEdit::OnDefaultEvent(wxMouseEvent& event)
{
  long now;
  float scrollx, scrolly;
  float x, y;
  Bool ateol;
  wxClickback *click;
  wxDC *dc;

  if (!admin)
    return;

  x = event.x;
  y = event.y;
  dc = admin->GetDC(&scrollx, &scrolly);
  y += scrolly;
  x += scrollx;

  if (!dc)
    return;

  now = FindPosition(x, y, &ateol);

  if (event.ButtonDown()) {
    if (click = FindClickback(now)) {
      tracking = TRUE;
      trackClickback = click;
      IBeam(FALSE);
      SetClickbackHilited(trackClickback, TRUE);
    } else {
      dragstart = now;
      dragging = TRUE;
      if (event.ShiftDown()) {
	if (dragstart > startpos)
	  dragstart = startpos;
	else
	  dragstart = endpos;
      }
      if (now < dragstart)
	SetPositionBiasScroll(-2, now, dragstart, ateol);
      else
	SetPositionBiasScroll(2, dragstart, now, ateol);
    }
  } else if (event.Dragging()) {
    now = FindPosition(x, y, &ateol);
    if (dragging) {
      if (now < dragstart) {
	if (now != startpos || dragstart != endpos)
	  SetPositionBiasScroll(-2, now, dragstart, ateol);
      } else {
	if (now != endpos || dragstart != startpos)
	  SetPositionBiasScroll(2, dragstart, now, ateol);
      }
    } else if (tracking) {
      SetClickbackHilited(trackClickback, now >= trackClickback->start
			  && now <= trackClickback->end);
    }
  } else if (event.ButtonUp()) {
    if (dragging)
      dragging = FALSE;
    else if (tracking) {
      if (trackClickback->hilited) {
	SetClickbackHilited(trackClickback, FALSE);
	trackClickback->f(this, trackClickback->start, 
			  trackClickback->end, trackClickback->data);
      }
      IBeam(TRUE);
    }
  } else if (event.Moving()) {
    now = FindPosition(x, y, &ateol);
    IBeam(!FindClickback(now));
  }
}

void wxMediaEdit::OnChar(wxKeyEvent &event)
{
  if (caretSnip && (caretSnip->flags & wxSNIP_HANDLES_EVENTS)) {
    wxDC *dc;
    float scrollx, scrolly, x, y;
    
    dc = admin->GetDC(&scrollx, &scrolly);
    GetSnipPositionAndLocation(caretSnip, NULL, &x, &y);

    caretSnip->OnChar(dc, x - scrollx, y - scrolly, x, y, event);
    return;
  }

  OnLocalChar(event);
}

void wxMediaEdit::OnDefaultChar(wxKeyEvent &event)
{
  long code;

  code = event.KeyCode();

  if (code == WXK_RETURN 
      || code == WXK_TAB
      || (code >= 32 && code <= 126)) {
    Insert(code);
  } else  {
    switch(code) {
    case WXK_BACK:
    case WXK_DELETE:
      Delete();
      break;
    case WXK_RIGHT:
    case WXK_LEFT:
    case WXK_UP:
    case WXK_DOWN:
    case WXK_HOME:
    case WXK_END:
      MovePosition(code, event.ShiftDown());
      break;
    }
  }
}

void wxMediaEdit::OwnCaret(Bool ownit)
{
  if (DoOwnCaret(ownit))
    NeedCaretRefresh();
}

void wxMediaEdit::SizeCacheInvalid(void)
{
  if (!graphicMaybeInvalid)
    graphicMaybeInvalid = TRUE;
  graphicsInvalid = TRUE;
  if (maxWidth > 0)
    flowInvalid = TRUE;
  snipCacheInvalid = TRUE;
}

/****************************************************************/

void wxMediaEdit::OnChange(void)
{
  /* Do nothing */
}

Bool wxMediaEdit::OnInsert(long start, long len)
{
  /* Do nothing */
  return TRUE;
}

void wxMediaEdit::AfterInsert(long start, long len)
{
}

Bool wxMediaEdit::OnDelete(long start, long len)
{
  /* Do nothing */
  return TRUE;
}

void wxMediaEdit::AfterDelete(long start, long len)
{
  /* Do nothing */
}

void wxMediaEdit::OnEditSequence(void)
{
  /* Do nothing */
}

void wxMediaEdit::AfterEditSequence(void)
{
  /* Do nothing */
}

/****************************************************************/

void wxMediaEdit::BeginEditSequence(Bool undoable)
{
  EndStreaks(wxSTREAK_EXCEPT_DELAYED);
  if (!delayRefresh) {
    if (!undoable)
      noundomode = TRUE;
    OnEditSequence();
  }
  delayRefresh++;
}

void wxMediaEdit::EndEditSequence(void)
{
  if (!(--delayRefresh)) {
    EndStreaks();
    Redraw();
    noundomode = FALSE;
    AfterEditSequence();
  } else if (delayRefresh < 0)
    delayRefresh = 0;
}

void wxMediaEdit::Recalculate(void)
{
}

void wxMediaEdit::GetPosition(long *start, long *end)
{
  if (start)
    *start = startpos;
  if (end)
    *end = endpos;
}

long wxMediaEdit::GetStartPosition(void)
{
  return startpos;
}

long wxMediaEdit::GetEndPosition(void)
{
  return endpos;
}
  
void wxMediaEdit::SetPosition(long start, long end, Bool ateol, Bool scroll)
{
  _SetPosition(FALSE, 0, start, end, ateol, scroll);
}

void wxMediaEdit::SetPositionBiasScroll(int bias, long start, long end, 
					Bool ateol, Bool scroll)
{
  _SetPosition(FALSE, bias, start, end, ateol, scroll);
}

Bool wxMediaEdit::ScrollToPosition(long start, Bool ateol, Bool refresh,
				   long end, int bias)
{
  float topx, botx, topy, boty;

  if (locked)
    return FALSE;

  if (end < start)
    end = start;

  if (delayRefresh) {
    delayedscrollbox = FALSE;
    delayedscroll = start;
    delayedscrollend = end;
    delayedscrollateol = ateol;
    return FALSE;
  }

  if (!CheckRecalc())
    return FALSE;

  delayedscroll = -1;
  
  PositionLocation(start, &topx, &topy, TRUE, ateol, TRUE);
  PositionLocation(end, &botx, &boty, FALSE, ateol, TRUE);

  return admin->ScrollTo(topx, topy, botx - topx, boty - topy, refresh, bias);
}

Bool wxMediaEdit::ScrollToPosition(long start, Bool ateol, long end, int bias)
{
  return ScrollToPosition(start, ateol, TRUE, end, bias);
}

void wxMediaEdit::GetVisiblePositionRange(long *start, long *end)
{
  float x, y, h, w;

  if (!CheckRecalc())
    return;

  admin->GetView(&x, &y, &w, &h);
  if (start)
    *start = FindPosition(x, y);
  if (end)
    *end = FindPosition(x + w, y + h);
}

void wxMediaEdit::GetVisibleLineRange(long *start, long *end)
{  
  float x, y, h, w;

  if (!CheckRecalc())
    return;

  admin->GetView(&x, &y, &w, &h);
  if (start)
    *start = FindLine(y);
  if (end)
    *end = FindLine(y + h);
}

void wxMediaEdit::MovePosition(long code, Bool extendSelection,
			       int kind)
{
  long i, start, end, extendstart, extendend;
  Bool ateol, vcursor, anchor, extend;
  float h, w, y;

  if (!CheckRecalc(maxWidth > 0))
    return;

  anchor = anchorStreak;
  vcursor = vcursorStreak;

  EndStreaks(wxSTREAK_EXCEPT_DELAYED);

  extendend = endpos;
  extendstart = startpos;
  extend = anchor || extendSelection;

  if (code == WXK_HOME)
    SetPositionBiasScroll(-1, 0, extend ? extendend : 0);
  else if (code == WXK_END)
    SetPositionBiasScroll(1, extend ? extendstart : len, len);
  else if (code == WXK_LEFT) {
    if (kind == wxMOVE_WORD) {
      start = startpos;
      FindWordbreak(&start, NULL, wxBREAK_FOR_CARET);
      end = (extend ? extendend : start);
      SetPositionBiasScroll(-1, start, end);
    } else if (kind == wxMOVE_LINE) {
      start = LineStartPosition(PositionLine(startpos, posateol));
      end = extend ? extendend : start;
      SetPositionBiasScroll(-1, start, end);
    } else {
      if (!extend && startpos == endpos)
	SetPositionBiasScroll(-1, startpos - 1);
      else if (extend)
	SetPositionBiasScroll(-1, startpos - 1, extendend);
      else
	SetPosition(startpos);
    }
  } else if (code == WXK_RIGHT) {
    if (kind == wxMOVE_WORD) {
      end = endpos;
      FindWordbreak(NULL, &end, wxBREAK_FOR_CARET);
      start = (extend ? extendstart : end);
      SetPositionBiasScroll(1, start, end);
    } else if (kind == wxMOVE_LINE) {
      end = LineEndPosition(PositionLine(startpos, posateol));
      start = extend ? extendstart : end;
      SetPositionBiasScroll(1, start, end);
    } else {
      if (!extend && startpos == endpos)
	SetPositionBiasScroll(1, startpos + 1, startpos + 1, TRUE);
      else if (extend)
	SetPositionBiasScroll(1, extendstart, endpos + 1, TRUE);
      else
	SetPosition(endpos, endpos, TRUE);
    }
  } else if ((code == WXK_UP) || (code == WXK_DOWN)) {
    if (!vcursor) {
      PositionLocation(startpos, &startloc, NULL, TRUE, posateol, TRUE);
      PositionLocation(endpos, &endloc, NULL, TRUE, posateol, TRUE);
    }
    if (code == WXK_UP) {
      if (kind == wxMOVE_PAGE) {
	admin->GetView(NULL, NULL, &w, &h);
	PositionLocation(startpos, NULL, &y, TRUE, posateol, TRUE);
	i = FindLine(y - h);
      } else
	i = PositionLine(startpos, posateol) - 1;
      if (i >= 0)
	start = FindPositionInLine(i, startloc, &ateol);
      else {
	start = 0;
	ateol = FALSE;
      }
      if (!extend)
	end = start;
      else
	end = extendend;
    } else {
      if (kind == wxMOVE_PAGE) {
	admin->GetView(NULL, NULL, &w, &h);
	PositionLocation(startpos, NULL, &y, TRUE, posateol, TRUE);
	i = FindLine(y + h);
      } else
	i = PositionLine(endpos, posateol) + 1;
      if (i <= numValidLines - 1)
	end = FindPositionInLine(i, endloc, &ateol);
      else
	end = len;
      if (!extend)
	start = end;
      else
	start = extendstart;
    }
    SetPositionBiasScroll(code == WXK_UP, start, end, ateol, TRUE);
    vcursorStreak = TRUE;
  }

  if (anchor)
    anchorStreak = TRUE;
}

void wxMediaEdit::SetAnchor(Bool on)
{
  anchorStreak = on;
}

Bool wxMediaEdit::GetAnchor(void)
{
  return anchorStreak;
}

void wxMediaEdit::_Insert(wxSnip *isnip, long strlen, char *str, 
			  long start, long end)
{
  long addlen, i, sPos, s;
  wxSnip *gsnip, *cSnip;
  wxTextSnip *snip;
  wxTabSnip *tabsnip;
  Bool deleted = FALSE, insertedLine = FALSE, scroll;

  if (locked || userLocked || (!isnip && !strlen))
    return;

  if (start < 0)
    return;
  if (start > len)
    start = len;

  if (caretStyle)
    if (start != end || start != startpos)
      caretStyle = NULL;

  if (end > -1)
    if (start < end) {
      delayRefresh++;
      Delete(start, end);
      deleted = TRUE;
      --delayRefresh;
    }

  if (isnip) {
    if (!isnip->count)
      return;

    if (!isnip->snipclass)
      wxMessageBox("Inserting a snip without a class."
		   " Data will be lost if you try to save the file.", 
		   "Warning");

    addlen = isnip->count;

    if (isnip->IsOwned())
      return;

    if (!OnInsert(start, addlen))
      return;

    if ((isnip->flags & wxSNIP_NEWLINE)&&!(isnip->flags & wxSNIP_HARD_NEWLINE))
      isnip->flags -= wxSNIP_NEWLINE;
    
    if (!len) {
      /* Special case: ignore the empty snip */
      snips = lastSnip = isnip;
      isnip->line = lineRoot = new wxMediaLine;
      lineRoot->snip = lineRoot->lastSnip = isnip;
      if (maxWidth > 0)
	lineRoot->MarkCheckFlow();
    } else {
      MakeSnipset(start, start);
      gsnip = FindSnip(start, +2);
      
      if (!gsnip) {
	AppendSnip(isnip);
	gsnip = lastLine->lastSnip;
	if (gsnip && (gsnip->flags & wxSNIP_HARD_NEWLINE)) {
	  isnip->line = lastLine->Insert(&lineRoot, FALSE);
	  isnip->line->snip = isnip->line->lastSnip = isnip;
	  numValidLines++;
	  insertedLine = TRUE;
	} else {
	  isnip->line = lastLine;
	  if (!lastLine->snip)
	    lastLine->snip = isnip;
	  lastLine->lastSnip = isnip;
	  if (isnip->flags & wxSNIP_HARD_NEWLINE)
	    insertedLine = TRUE; /* b/c added extra ghost line */
	}
      } else {
	InsertSnip(gsnip, isnip);
	if (isnip->flags & wxSNIP_HARD_NEWLINE) {
	  isnip->line = gsnip->line->Insert(&lineRoot, TRUE);
	  insertedLine = TRUE;
	  numValidLines++;
	  if (PTREQ(gsnip->line->snip, gsnip))
	    isnip->line->snip = isnip;
	  else
	    isnip->line->snip = gsnip->line->snip;
	  isnip->line->lastSnip = isnip;
	  gsnip->line->snip = gsnip;
	  
	  for (cSnip = isnip->line->snip; PTRNE(cSnip, isnip); 
	       cSnip = cSnip->next)
	    cSnip->line = isnip->line;
	  
	  gsnip->line->CalcLineLength();
	  gsnip->line->MarkRecalculate();
	} else {
	  isnip->line = gsnip->line;
	  if (PTREQ(isnip->line->snip, gsnip))
	    isnip->line->snip = isnip;
	}
      }

      if (maxWidth > 0) {
	isnip->line->MarkCheckFlow();
	if (isnip->line->prev
	    && !(isnip->line->prev->lastSnip->flags & wxSNIP_HARD_NEWLINE))
	  isnip->line->prev->MarkCheckFlow();
	if ((isnip->flags & wxSNIP_HARD_NEWLINE) && isnip->line->next)
	  isnip->line->next->MarkCheckFlow();
      }
    }

    isnip->SizeCacheInvalid();
    isnip->style = styleList->Convert(isnip->style);  

    isnip->line->CalcLineLength();
    isnip->line->MarkRecalculate();
    
    isnip->SetAdmin(snipAdmin);
  } else {
    addlen = strlen;
    
    if (!OnInsert(start, addlen))
      return;

    if (!len) {
      snip = InsertTextSnip(start, caretStyle);
      sPos = 0;
      caretStyle = NULL;
      lineRoot->snip = lineRoot->lastSnip = snip;
    } else {
      if (start)
	gsnip = FindSnip(start, -1, &sPos);
      else
	gsnip = NULL;

      if (!gsnip || caretStyle || !(gsnip->flags & wxSNIP_IS_TEXT)) {
	snip = InsertTextSnip(start, caretStyle);
	caretStyle = NULL;
	sPos = start;
      } else {
	snip = (wxTextSnip *)gsnip;
	if (!(snip->flags & wxSNIP_CAN_APPEND)) {
	  snip = InsertTextSnip(start, snip->style);
	  sPos = start;
	}
      }

      if (gsnip && gsnip->flags & wxSNIP_HARD_NEWLINE) {
	wxMediaLine *oldline = gsnip->line, *newline;
	
	if (!oldline->next) {
	  oldline->Insert(&lineRoot, FALSE);
	  insertedLine = TRUE;
	  numValidLines++;
	  
	  oldline->next->lastSnip = snip;
	}
      
	newline = oldline->next;
	
	snip->line = newline;
	
	oldline->lastSnip = gsnip;
	newline->snip = snip;
	
	oldline->CalcLineLength();
	oldline->MarkRecalculate();
      }
    }

    s = start - sPos;
    snip->Insert(str, addlen, s);

    snip->line->CalcLineLength();
    snip->line->MarkRecalculate();

    if (maxWidth > 0) {
      snip->line->MarkCheckFlow();
      if (snip->line->prev 
	  && !(snip->line->prev->lastSnip->flags & wxSNIP_HARD_NEWLINE))
	snip->line->prev->MarkCheckFlow();
    }

    str = snip->text + s;
    for (i = 0; i < addlen; i++) {
      if (*str == '\r')
	*str = '\n';
      if (*str == '\n' || *str == '\t') {
	Bool newline = (*str == '\n');

	MakeSnipset(i + start, i + start + 1);
	snip = (wxTextSnip *)FindSnip(i + start, +1);

	if (newline) {
	  /* Forced return - split the snip */
	  wxMediaLine *oldLine;

	  snip->flags |= wxSNIP_NEWLINE | wxSNIP_HARD_NEWLINE 
	    | wxSNIP_INVISIBLE;
	  snip->flags -= (snip->flags & wxSNIP_CAN_APPEND);

	  if (PTRNE(snip, snip->line->lastSnip)) {
	    oldLine = snip->line;
	    snip->line = oldLine->Insert(&lineRoot, TRUE);
	    insertedLine = TRUE;
	    numValidLines++;
	    snip->line->lastSnip = snip;
	    snip->line->snip = oldLine->snip;
	    
	    /* Retarget snips moved to new line: */
	    for (cSnip = snip->line->snip; PTRNE(cSnip, snip);
		 cSnip = cSnip->next)
	      cSnip->line = snip->line;

	    oldLine->snip = snip->next;

	    oldLine->CalcLineLength();
	    oldLine->MarkRecalculate();
	    if (maxWidth > 0)
	      oldLine->MarkCheckFlow();

	    snip->line->CalcLineLength();
	    snip->line->MarkRecalculate();
	    if (maxWidth > 0)
	      snip->line->MarkCheckFlow();
	  } else if (PTREQ(snip, lastSnip))
	    insertedLine = TRUE; /* b/c added extra ghost line */
	} else {
	  tabsnip = new wxTabSnip();
	  tabsnip->style = snip->style;
	  tabsnip->SetAdmin(snipAdmin);
	  tabsnip->Insert("\t", 1, 0);
	  SpliceSnip(tabsnip, snip->prev, snip->next);
	  tabsnip->line = snip->line;
	  if (PTREQ(snip->line->snip, snip))
	    tabsnip->line->snip = tabsnip;
	  if (PTREQ(snip->line->lastSnip, snip))
	    tabsnip->line->lastSnip = tabsnip;
	  delete snip;
	}

	snip = (wxTextSnip *)FindSnip(i + start + 1, +1);
	str = snip->text;
      } else
	str++;
    }
  }

  firstLine = lineRoot->First();
  lastLine = lineRoot->Last();

  len += addlen;
  
  AdjustClickbacks(start, start, addlen, NULL);

  if (!modified)
    AddUndo(new wxUnmodifyRecord);
  AddUndo(new wxInsertRecord(start, addlen, 
			     deleted || typingStreak || delayedStreak
			     || !modified));
  if (delayRefresh)
    delayedStreak = TRUE;

  scroll = (startpos == start);

  if (startpos >= start)
    startpos += addlen;
  if (endpos >= start)
    endpos += addlen;

  extraLine = !!(lastSnip->flags & wxSNIP_NEWLINE);

  if (scroll) {
    delayRefresh++;
    ScrollToPosition(startpos);
    --delayRefresh;
  }

  changed = TRUE;

  caretStyle = NULL;

  if (insertedLine)
    NeedRefresh(start);
  else
    RefreshByLineDemand();

  if (!modified)
    SetModified(TRUE);

  AfterInsert(start, addlen);
}

void wxMediaEdit::Insert(char *str, long start, long end)
{
  _Insert(NULL, strlen(str), str, start, end);
}

void wxMediaEdit::Insert(char *str)
{
  Insert(str, startpos, endpos);
}

void wxMediaEdit::Insert(long len, char *str)
{
  _Insert(NULL, len, str, startpos, endpos);
}

void wxMediaEdit::Insert(unsigned char ascii, long start, long end)
{
  Bool streak;
  unsigned char buffer[2];

  buffer[0] = ascii;
  buffer[1] = 0;
  
  streak = typingStreak;
  EndStreaks();
  typingStreak = streak;

  Insert((char *)buffer, start, end);

  typingStreak = TRUE;
}

void wxMediaEdit::Insert(unsigned char ascii)
{
  Insert(ascii, startpos, endpos);
}

void wxMediaEdit::Insert(wxSnip *snip, long start, long end)
{
  _Insert(snip, 0, NULL, start, end);
}

void wxMediaEdit::Insert(wxSnip *snip)
{
  Insert(snip, startpos, endpos);
}

void wxMediaEdit::_Delete(long start, long end, Bool withUndo)
{
  wxSnip *prev, *next, *startSnip, *endSnip;
  long dellen;
  wxSnip *snip;
  wxDeleteRecord *rec;
  wxMediaLine *line;
  Bool deletedLine = FALSE;

  if (locked || userLocked)
    return;

  if (end == -1) {
    if (!start)
      return;
    end = start;
    --start;
  }

  if (start >= end)
    return;
  if ((start < 0) || (start >= len))
    return;

  if (end > len)
    end = len;
  
  if (!OnDelete(start, end - start))
    return;

  MakeSnipset(start, end);
  
  startSnip = FindSnip(start, -2);
  endSnip = FindSnip(end, -1);

  if (withUndo) {
    if (!modified)
      AddUndo(new wxUnmodifyRecord);
    rec = new wxDeleteRecord(start, end, deletionStreak || delayedStreak
			     || !modified);
  }

  for (snip = endSnip; PTRNE(snip, startSnip); snip = prev) {
    if (PTREQ(snip, caretSnip)) {
      caretSnip->OwnCaret(FALSE);
      caretSnip = NULL;
    }

    if (withUndo)
      rec->InsertSnip(snip);

    prev = snip->prev;

    if (PTREQ(snip->line->snip, snip)) {
      if (PTREQ(snip->line->lastSnip, snip)) {
	snip->line->Delete(&lineRoot);
	deletedLine = TRUE;
	numValidLines--;
      } else
	snip->line->snip = snip->next;
    } else if (PTREQ(snip->line->lastSnip, snip)) {
      if (snip->line->next) {
	snip->line->lastSnip = snip->line->next->lastSnip;
	snip->line->next->Delete(&lineRoot);
	deletedLine = TRUE;
	numValidLines--;
      } else {
	snip->line->lastSnip = prev;
	if (!snip->line->next && extraLine)
	  deletedLine = TRUE; /* b/c maybe deleted extra ghost line */
      }
    }

    DeleteSnip(snip);
  }

  if (!snipCount)
    MakeOnlySnip();

  firstLine = lineRoot->First();
  lastLine = lineRoot->Last();

  if (startSnip) {
    if (startSnip->flags & wxSNIP_NEWLINE) {
      if (startSnip->line->next)
	line = startSnip->line->next;
      else
	line = NULL;
    } else
      line = startSnip->line;
  } else
    line = firstLine;

  if (line) {
    /* Fix line references from possibly moved snips: */
    next = line->lastSnip->next;
    for (snip = line->snip; PTRNE(snip, next); snip = snip->next)
      snip->line = line;
    
    line->CalcLineLength();
    line->MarkRecalculate();

    if (maxWidth >= 0) {
      line->MarkCheckFlow();
      if (line->prev && !(line->prev->lastSnip->flags & wxSNIP_HARD_NEWLINE))
	line->prev->MarkCheckFlow();
    }
  }

  AdjustClickbacks(start, end, start - end, rec);

  if (withUndo) {
    AddUndo(rec);
    if (delayRefresh)
      delayedStreak = TRUE;
  }

  dellen = end - start;
  
  len -= dellen;

  CheckMergeSnips(start);

  if ((startpos >= start) && (startpos <= end))
    startpos = start;
  else if (startpos > end)
    startpos -= dellen;

  if ((endpos >= start) && (endpos <= end))
    endpos = start;
  else if (endpos > end)
    endpos -= dellen;

  extraLine = !!(lastSnip->flags & wxSNIP_NEWLINE);

  if (startpos == start) {
    delayRefresh++;
    ScrollToPosition(startpos);
    --delayRefresh;
  }

  changed = TRUE;

  caretStyle = NULL;

  if (start == len) {
    /* force recheck extra line state: */
    graphicMaybeInvalid = -1;
  }

  if (deletedLine)
    NeedRefresh(start);
  else
    RefreshByLineDemand();

  if (!modified)
    SetModified(TRUE);

  AfterDelete(start, dellen);
}

void wxMediaEdit::Delete(long start, long end)
{
  _Delete(start, end, TRUE);
}

void wxMediaEdit::Delete()
{
  Bool dstreak, streak = (startpos == endpos);

  dstreak = deletionStreak;
  EndStreaks();
  deletionStreak = dstreak;
  Delete(startpos, (endpos == startpos) ? -1 : endpos);
  deletionStreak = streak;
}

void wxMediaEdit::Erase(void)
{
  Delete(0, len);
}

void wxMediaEdit::Clear()
{
  Delete(startpos, endpos);
}

void wxMediaEdit::Cut(Bool extend, long time, long start, long end)
{
  if (start == -1)
    start = startpos;
  if (end == -1)
    end = endpos;
  if (start < 0)
    start = 0;
  if (end > len)
    end = len;
  if (start >= end)
    return;
  Copy(extend, time, start, end);
  Delete(start, end);
}

void wxMediaEdit::Cut(Bool extend, long time)
{
  Cut(extend, time, -1);
}

void wxMediaEdit::Copy(Bool extend, long time, long startp, long endp)
{
  wxSnip *start, *end;
  wxSnip *asnip, *snip;

  if (startp == -1)
    startp = startpos;
  if (endp == -1)
    endp = endpos;
  if (startp < 0)
    startp = 0;
  if (endp > len)
    endp = len;
  if (startp >= endp)
    return;

  if (!extend)
    FreeOldCopies(); 
  
  MakeSnipset(startp, endp);
  
  wxmb_commonCopyRegionData = GetRegionData(startp, endp);

  start = FindSnip(startp, +1);
  end = FindSnip(endp, +2);

  for (snip = start; PTRNE(snip, end); snip = snip->next) {
    asnip = snip->Copy();
    asnip->SetAdmin(NULL);
    wxmb_commonCopyBuffer->Append(asnip);
    wxmb_commonCopyBuffer2->Append(GetSnipData(snip));
  }

  InstallCopyBuffer(time);
}

void wxMediaEdit::Copy(Bool extend, long time)
{
  Copy(extend, time, -1);
}

void wxMediaEdit::Paste(long time, long start, long end)
{
  if (end == -1)
    end = (start == -1) ? endpos : start;
  if (start == -1)
    start = endpos;
  if (start < 0)
    start = 0;
  if (end > len)
    end = len;
  if (start > end)
    return;

  BeginEditSequence();

  if (start < end)
    Delete(start, end);

  readInsert = readInsertStart = start;

  DoPaste(time);

  EndEditSequence();
}

void wxMediaEdit::InsertPasteSnip(wxSnip *snip, wxBufferData *data)
{
  Insert(snip, readInsert);
  SetSnipData(snip, data);
  readInsert += snip->count;
}

void wxMediaEdit::PasteRegionData(wxBufferData *data)
{
  SetRegionData(readInsertStart, readInsert, data);
}

void wxMediaEdit::InsertPasteString(char *str)
{
  Insert(str, readInsert);
  readInsert += strlen(str);
}

void wxMediaEdit::Paste(long time)
{
  Paste(time, -1);
}

void wxMediaEdit::Kill(long time, long start, long end)
{
  int streak;
  char *text;

  if ((start < 0) != (end < 0))
    return;

  streak = killStreak;

  BeginEditSequence();
  if (start < 0) {
    MovePosition(WXK_RIGHT, TRUE, wxMOVE_LINE);
    if (startpos == endpos)
      SetPosition(startpos, startpos + 1);
    else {
      long i;
      
      text = GetText(startpos, endpos);
      for (i = endpos - startpos; i--; )
	if (!isspace(text[i]))
	  break;
      
      if (i < 0) {
	/* Line has all spaces: move one more */
	SetPosition(startpos, endpos + 1);
      }
      
      delete[] text;
    }
    start = startpos;
    end = endpos;
  }
  Cut(streak, time, start, end);
  EndEditSequence();

  killStreak = TRUE;
}

void wxMediaEdit::Kill(long time)
{
  Kill(time, -1, -1);
}

/****************************************************************/

char *wxMediaEdit::GetFlattenedText(void)
{
  return GetText(-1, -1, TRUE);
}

char *wxMediaEdit::GetText(long start, long end, Bool flatt, Bool forceCR)
{
  wxSnip *snip;
  long count, sPos, p, num, offset, total;
  long alloc;
  char *s, *t, *old;

  if (start < 0)
    start = 0;
  if (end < 0)
    end = len;

  if (start > len)
    start = len;
  if (end < start)
    end = start;

  count = end - start;

  if (!flatt) {
    s = new char[count + 1];
    s[count] = 0;
  } else {
    alloc = 2 * count;
    if (!alloc)
      alloc = 2;
    s = new char[alloc];
    s[0] = 0;
  }

  if (!count)
    return s;

  snip = FindSnip(start, +1, &sPos);
  offset = start - sPos;
  num = ((snip->count - offset <= count) ? snip->count - offset : count);
  if (!flatt) {
    snip->GetText(s, offset, num);
    p = num;
  } else {
    int add_newline;

    t = snip->GetText(offset, num, TRUE);
    p = strlen(t);
    if (forceCR && (snip->flags & wxSNIP_NEWLINE) 
	  && !(snip->flags & wxSNIP_HARD_NEWLINE)) {
      p++;
      add_newline = 1;
    } else
      add_newline = 0;
    if (p >= alloc) {
      alloc = 2 * p;
      delete[] s;
      s = new char[alloc];
    }
    memcpy(s, t, p - add_newline);
    if (add_newline)
      s[p - 1] = '\n';
    delete[] t;
  }
  total = num;
  snip = snip->next;
  while (snip && (count > total)) {
    num = ((total + snip->count <= count) ? snip->count : count - total);
    if (!flatt) {
      snip->GetText(s + p, 0, num);
      p += num;
    } else {
      int add_newline;

      t = snip->GetText(0, num, TRUE);
      offset = strlen(t);

      if (forceCR && (snip->flags & wxSNIP_NEWLINE) 
	  && !(snip->flags & wxSNIP_HARD_NEWLINE)) {
	offset++;
	add_newline = 1;
      } else
	add_newline = 0;

      if (p + offset >= alloc) {
	alloc = 2 * (p + offset);
	old = s;
	s = new char[alloc];
	memcpy(s, old, p);
	delete[] old;
      }

      memcpy(s + p, t, offset);
      if (add_newline)
	s[p + offset - 1] = '\n';
      p += offset;

      delete[] t;
    }
    total += num;
    snip = snip->next;
  }

  if (flatt)
    s[p] = 0;

  return s;
}

unsigned char wxMediaEdit::GetCharacter(long start)
{
  wxSnip *snip;
  long sPos;
  unsigned char buffer[2];

  if (start < 0)
    start = 0;
  else if (start >= len)
    return 0;

  snip = FindSnip(start, +1, &sPos);
  snip->GetText((char *)buffer, start - sPos, 1);

  return buffer[0];
}

/****************************************************************/

void wxMediaEdit::SetClickback(long start, long end, 
			       wxClickbackFunc f, void *d,
			       wxStyleDelta *delta)
{
  wxClickback *click;

  click = new wxClickback;
  click->start = start;
  click->end = end;
  click->f = f;
  click->data = d;
  WXGC_IGNORE(click->data);

  click->delta = new wxStyleDelta;
  if (delta)
    memcpy(click->delta, delta, sizeof(wxStyleDelta));

  SetClickback(click);
} 

void wxMediaEdit::SetClickback(wxClickback *click)
{
  clickbacks->Append(click);
}

void wxMediaEdit::RemoveClickback(long start, long end)
{
  wxNode *node, *next;
  wxClickback *click;

  for (node = clickbacks->First(); node; node = next) {
    next = node->Next();
    click = (wxClickback *)node->Data();
    if (click->start == start && click->end == end) {
      delete click;
      clickbacks->DeleteNode(node);
    }
  }
}

/****************************************************************/

void wxMediaEdit::FlashOn(long start, long end, Bool ateol, Bool scroll, 
			  long timeout)
{
  _SetPosition(TRUE, 0, start, end, ateol, scroll);
  if (timeout > 0) {
    flashautoreset = TRUE;
    flashTimer = new wxMediaFlashTimer();
    flashTimer->media = this;
    WXGC_IGNORE(flashTimer->media);
    flashTimer->Start(timeout);
  }
  flashscroll = scroll;
}

void wxMediaEdit::FlashOff(void)
{
  if (!flash)
    return;
  flashautoreset = -1;
  _SetPosition(FALSE, 0, startpos, endpos, posateol, flashscroll);
}

/****************************************************************/

void StandardWordbreak(wxMediaEdit *win, long *startp, long *endp, 
		       int reason, void *)
{
  long pstart, start, lstart, end, lend;
  unsigned char *text;
  char *map;
  wxMediaWordbreakMap *wordBreakMap;

  wordBreakMap = win->GetWordbreakMap();

  map = wordBreakMap ? wordBreakMap->map : wxTheMediaWordbreakMap.map;

#define nonbreak(x) (map[x] & reason)

  if (startp) {
    pstart = start = *startp;

    lstart = win->FindNewline(-1, start);
    if (lstart < 0)
      lstart = 0;
    else if (reason == wxBREAK_FOR_CARET) {
      lstart = win->FindNewline(-1, lstart - 1);
      if (lstart < 0)
	lstart = 0;
    }
    lend = win->FindNewline(1, start);
    if (lend < 0)
      lend = win->LastPosition();

    text = (unsigned char *)win->GetText(lstart, lend);
    
    start -= lstart;
    if (start && nonbreak(text[start]))
      --start;
    if (reason != wxBREAK_FOR_SELECTION) {
      while (start && !nonbreak(text[start]))
	--start;
    }
    while (start && nonbreak(text[start]))
      --start;
    if ((start < pstart) && !nonbreak(text[start]))
      start++;

    *startp = start + lstart;

    delete[] text;
  }

  if (endp) {
    end = *endp;

    lstart = win->FindNewline(-1, end);
    if (lstart < 0)
      lstart = 0;
    lend = win->FindNewline(1, end);
    if (lend < 0)
      lend = win->LastPosition();
    else if (reason == wxBREAK_FOR_CARET) {
      lend = win->FindNewline(1, lend + 1);
      if (lend < 0)
	lend = win->LastPosition();
    }
    text = (unsigned char *)win->GetText(lstart, lend);
    
    end -= lstart;
    lend -= lstart;
    while ((end < lend) && !nonbreak(text[end]))
      end++;
    while ((end < lend) && nonbreak(text[end]))
      end++;

    *endp = end + lstart;

    delete[] text;
  }
}

/****************************************************************/

void wxMediaEdit::SetWordbreakFunc(wxWordbreakFunc f, void *data)
{
  wordBreak = f;
  wordBreakData = data;
  WXGC_IGNORE(wordBreakData);
}

void wxMediaEdit::FindWordbreak(long *start, long *end, int reason)
{
  long oldstart, oldend;

  if (start)
    oldstart = *start;
  if (end)
    oldend = *end;

  wordBreak(this, start, end, reason, wordBreakData);

  if (start && *start > oldstart)
    *start = oldstart;
  if (end && *end < oldend)
    *end = oldend;
}

wxMediaWordbreakMap *wxMediaEdit::GetWordbreakMap(void)
{
  return wordBreakMap;
}

void wxMediaEdit::SetWordbreakMap(wxMediaWordbreakMap *map)
{
  if (wordBreakMap) {
    wordBreakMap->AdjustUsage(FALSE);
#if !WXGARBAGE_COLLECTION_ON
    if (!wordBreakMap->IsUsed())
      delete wordBreakMap;
#endif
  }

  wordBreakMap = map;

  if (!map)
    return;

  map->AdjustUsage(TRUE);
}

/****************************************************************/

float wxMediaEdit::GetMaxWidth()
{
  return maxWidth;
}

float wxMediaEdit::GetMinWidth()
{
  return minWidth;
}

void wxMediaEdit::SetMaxWidth(float w)
{
  if (w == maxWidth)
    return;

  if (w > 0 && w < CURSOR_WIDTH)
    w = CURSOR_WIDTH + 1;
  maxWidth = w;
  flowInvalid = TRUE;
  if (!graphicMaybeInvalid)
    graphicMaybeInvalid = TRUE;
  NeedRefresh(-1, -1);
}

void wxMediaEdit::SetMinWidth(float w)
{
  if (w == minWidth)
    return;

  graphicMaybeInvalid = -1;
  minWidth = w;
  NeedRefresh(-1,-1);
}

void wxMediaEdit::SetMinHeight(float h)
{
  if (h == minHeight)
    return;

  graphicMaybeInvalid = -1;
  minHeight = h;
  NeedRefresh(-1,-1);
}

void wxMediaEdit::SetMaxHeight(float h)
{
  if (h == maxHeight)
    return;

  graphicMaybeInvalid = -1;
  maxHeight = h;
  NeedRefresh(-1,-1);
}

float wxMediaEdit::GetMinHeight()
{
  return minHeight;
}

float wxMediaEdit::GetMaxHeight()
{
  return maxHeight;
}

/****************************************************************/

Bool wxMediaEdit::LoadFile(char *file, int format)
{
  if (!file || !*file) {
    if ((file && !*file) || !filename || tempFilename) {
      char *path;
      
      if (filename)
	path = copystring(PathOnly(filename));
      else
	path = NULL;
      
      file = wxFileSelector("Choose a file", path, NULL,
			    NULL, NULL, wxOPEN);
      
      if (path)
	delete[] path;
    } else
      file = filename;
  }

  if (!file)
    return FALSE;

  if (::DirExists(file)) {
    wxMessageBox("Can't load a directory.", "Error");
    return FALSE;
  }

  FILE *f = fopen(file, "rb");
  
  if (!f) {
    wxMessageBox("Couldn't open the file.", "Error");
    return FALSE;
  }

  const int BUF_SIZE = 1000;
  char buffer[BUF_SIZE];
  long n;

  wxBeginBusyCursor();

  BeginEditSequence();

  Erase();

  if (PTRNE(file, filename))
    SetFilename(file, FALSE);
  
  if (format == wxMEDIA_FF_SAME)
    format = fileFormat;

  if (format == wxMEDIA_FF_GUESS) {
    n = fread((char *)buffer, 1, MRED_START_STR_LEN, f);
    buffer[MRED_START_STR_LEN] = 0;
    if ((n != MRED_START_STR_LEN) || strcmp(buffer, MRED_START_STR))
      format = wxMEDIA_FF_TEXT;
    else
      format = wxMEDIA_FF_STD;
    fseek(f, 0, 0);
  }

  if (format == wxMEDIA_FF_TEXT || format == wxMEDIA_FF_TEXT_FORCE_CR) {
    if (fclose(f))
      wxMessageBox("There was an error closing the file.", "Warning");
    f = fopen(file, "r");
    if (!f) {
      wxMessageBox("Couldn't open the file.", "Error");
      return FALSE;
    }
  }

  Bool fileerr = FALSE;

  if (format == wxMEDIA_FF_STD) {
    n = fread((char *)buffer, 1, MRED_START_STR_LEN, f);
    buffer[MRED_START_STR_LEN] = 0;
    if ((n != MRED_START_STR_LEN) || strcmp(buffer, MRED_START_STR)){
      wxMessageBox("This is not a MrEd file.", "Error");
      fseek(f, 0, 0);
      format = wxMEDIA_FF_TEXT;
    } else {
      fread((char *)wxme_current_read_format, 1, MRED_FORMAT_STR_LEN, f);
      fread((char *)wxme_current_read_version, 1, MRED_VERSION_STR_LEN, f);
      
      if (wxmeCheckFormatAndVersion()) {
	wxMediaStreamInFileBase b(f);
	wxMediaStreamIn mf(b);
	
	styleList->Clear();
	
	if (wxmbReadGlobalHeader(styleList, mf)) {
	  if (mf.Ok())
	    fileerr = !ReadFromFile(mf);
	  else
	    fileerr = TRUE;
	} else
	  fileerr = TRUE;
	fileerr = !wxmbReadGlobalFooter(mf) || fileerr;
	
	styleList->NewNamedStyle(STD_STYLE, NULL);
	
	fileerr = fileerr || !mf.Ok();
      } else
	fileerr = TRUE;
    }
  }

  if (format == wxMEDIA_FF_TEXT || format == wxMEDIA_FF_TEXT_FORCE_CR) {
    while (!ferror(f) && !feof(f)) {
      n = fread((char *)buffer, 1, BUF_SIZE, f);
      Insert(n, (char *)buffer);
    }
  } 

  fileerr = fileerr || ferror(f);

  if (fclose(f))
    wxMessageBox("There was an error closing the file.", "Warning");

  if (fileerr)
    wxMessageBox("There was an error loading the file.", "Error");

  fileFormat = format;

  SetPosition(0, 0);

  EndEditSequence();

  ClearUndos();

  wxEndBusyCursor();

  SetModified(fileerr);

  return !fileerr;
}

Bool wxMediaEdit::SaveFile(char *file, int format)
{
  if (!file || !*file) {
    if ((file && !*file) || !filename || tempFilename) {
      char *path, *pfile;
      
      if (filename) {
	path = copystring(PathOnly(filename));
	pfile = copystring(FileNameFromPath(filename));
      } else
	path = pfile = NULL;
      
      file = wxFileSelector("Save file as", path,
			    pfile, NULL, NULL, wxSAVE);
      
      if (path) {
	delete[] path;
	delete[] pfile;
      }
    } else
      file = filename;
  }

  if (!file)
    return FALSE;

  if ((format == wxMEDIA_FF_SAME) || (format == wxMEDIA_FF_GUESS))
    format = fileFormat;

  FILE *f = fopen(file, 
		  ((format == wxMEDIA_FF_TEXT) 
		   || (format == wxMEDIA_FF_TEXT_FORCE_CR)) ? "w" : "wb");

  if (!f) {
    wxMessageBox("Couldn't write the file.", "Error");
    return FALSE;
  }

  wxBeginBusyCursor();

  Bool fileerr = FALSE;

  if (format == wxMEDIA_FF_TEXT || format == wxMEDIA_FF_TEXT_FORCE_CR) {
    char *s = GetText(-1, -1, TRUE, format == wxMEDIA_FF_TEXT_FORCE_CR);
    fwrite(s, 1, strlen(s), f);
    fileerr = ferror(f);
    if (fclose(f)) {
      wxMessageBox("There was an error closing the file.", "Error");
      fileerr = TRUE;
    }
    delete[] s;
  } else {
    fwrite(MRED_START_STR, 1, MRED_START_STR_LEN, f);
    fwrite(MRED_FORMAT_STR, 1, MRED_FORMAT_STR_LEN, f);
    fwrite(MRED_VERSION_STR, 1, MRED_VERSION_STR_LEN, f);    

    wxMediaStreamOutFileBase b(f);
    wxMediaStreamOut mf(b);

    wxmbWriteGlobalHeader(styleList, mf);
    if (mf.Ok())
      fileerr = !WriteToFile(mf);
    wxmbWriteGlobalFooter(mf);

    fileerr = fileerr || !mf.Ok();

    if (fclose(f)) {
      wxMessageBox("There was an error closing the file.", "Error");
      fileerr = TRUE;
    }
  }

  if (fileerr)
    wxMessageBox("There was an error writing the file.", "Error");

  if (PTRNE(file, filename))
    SetFilename(file, FALSE);
  fileFormat = format;

  wxEndBusyCursor();

  SetModified(fileerr);

  return !fileerr;
}

Bool wxMediaEdit::ReadFromFile(wxMediaStreamIn &f, long start)
{
  if (start < 0)
    start = startpos;

  readInsert = start;

  return ReadSnipsFromFile(f);
}

Bool wxMediaEdit::ReadFromFile(wxMediaStreamIn &f)
{
  return ReadFromFile(f, -1);
}

Bool wxMediaEdit::ReadInsert(wxSnip *snip)
{
  Insert(snip, readInsert);
  readInsert += snip->count;
  return TRUE;
}

Bool wxMediaEdit::WriteToFile(wxMediaStreamOut &f, long start, long end)
{
  wxSnip *startSnip, *endSnip;

  if (start < 0)
    start = 0;
  if (end < 0)
    end = len;
    
  startSnip = FindSnip(start, +1);
  endSnip = FindSnip(end, +2);

  if (!DoWriteHeadersFooters(f, TRUE))
    return FALSE;

  wxmbWriteSnipsToFile(f, styleList, NULL, startSnip, endSnip, NULL, this);
  
  if (!DoWriteHeadersFooters(f, FALSE))
    return FALSE;

  return TRUE;
}

Bool wxMediaEdit::WriteToFile(wxMediaStreamOut &f)
{
  return WriteToFile(f, -1);
}

int wxMediaEdit::GetFileFormat(void)
{
  return fileFormat;
}

void wxMediaEdit::SetFileFormat(int format)
{
  if (format == wxMEDIA_FF_STD
      || format == wxMEDIA_FF_TEXT
      || format == wxMEDIA_FF_TEXT_FORCE_CR)
    fileFormat = format;
}

void wxMediaEdit::SetFilename(char *name, Bool temp)
{
  wxSnip *snip;

  if (filename)
    delete[] filename;
  filename = copystring(name);
  tempFilename = temp;

  for (snip = snips; snip; snip = snip->next)
    if (snip->flags & wxSNIP_USES_BUFFER_PATH)
      snip->SetAdmin(snipAdmin);
}

wxBufferData *wxMediaEdit::GetRegionData(long start, long end)
{
  return NULL;
}

void wxMediaEdit::SetRegionData(long start, long end, wxBufferData *)
{
}

/****************************************************************/

float *wxMediaEdit::GetTabs(int *count, float *space, Bool *inUnits)
{
  if (count)
    *count = tabcount;

  if (space)
    *space = tabSpace;

  if (inUnits)
    *inUnits = tabSpaceInUnits;

  return tabs;
}

void wxMediaEdit::SetTabs(float *newtabs, int count, 
			  float tabWidth, Bool inUnits)
{
  delete tabs;
  
  tabs = newtabs;
  tabcount = count;

  if (tabWidth >= 1)
    tabSpace = tabWidth;
  else
    tabSpace = wxTAB_WIDTH;

  tabSpaceInUnits = inUnits;

  SizeCacheInvalid();
  NeedRefresh(-1, -1);
}

/****************************************************************/

long wxMediaEdit::FindPositionInLine(long i, float x, Bool *ateol, Bool *onit)
{
  return _FindPositionInLine(FALSE, i, x, ateol, onit);
}

long wxMediaEdit::_FindPositionInSnip(wxDC *dc, float X, float Y,
				      wxSnip *snip, float x)
{
  long offset, range, i;

  if (x < 0)
    return 0;
  if (snip->PartialOffset(dc, X, Y, snip->count) <= x)
    return snip->count;

  /* Binary search for position within snip: */
  range = snip->count;
  i = range / 2;
  offset = 0;

  while (1) {
    if (snip->PartialOffset(dc, X, Y, offset + i) > x)
      range = i;
    else if (snip->PartialOffset(dc, X, Y, offset + i + 1) <= x) {
      offset += i;
      range -= i;
    } else
      break;
    
    i = range / 2;
  }

  return i + offset;
}

long wxMediaEdit::FindLine(float y, Bool *onit)
{
  if (onit)
    *onit = FALSE;
    
  if (!CheckRecalc())
    return 0;

  if (y <= 0)
    return 0;
  if ((y >= totalHeight) || (extraLine && y >= totalHeight - extraLineH))
    return numValidLines - (extraLine ? 0 : 1);

  if (onit)
    *onit = TRUE;

  return lineRoot->FindLocation(y)->GetLine();
}

long wxMediaEdit::FindPosition(float x, float y, Bool *ateol, Bool *onit)
{
  long i, p;
  Bool online;

  if (ateol)
    *ateol = FALSE;

  i = FindLine(y, &online);
  if ((i >= numValidLines - 1) && !online) {
    if (onit)
      *onit = FALSE;
    return len;
  }

  p = FindPositionInLine(i, x, ateol, onit);
  if (onit)
    *onit = online && *onit;

  return p;
}

long wxMediaEdit::PositionLine(long start, Bool eol)
{
  wxMediaLine *line;

  if (!CheckRecalc(maxWidth > 0))
    return 0;

  if (start <= 0)
    return 0;
  if (start >= len) {
    if (extraLine && !eol)
      return numValidLines;
    else
      return numValidLines - 1;
  }

  line = lineRoot->FindPosition(start);
  if (eol && (line->GetPosition() == start))
    line = line->prev;
  
  return line->GetLine();
}

void wxMediaEdit::PositionLocation(long start, float *x, float *y, 
				   Bool top, Bool eol, Bool wholeLine)
{
  float horiz, h, descent, space, topy;
  int align;
  wxMediaLine *line;
  wxSnip *snip;
  wxDC *dc;

  if (!CheckRecalc(TRUE))
    return;

  /* Handle boundary cases first: */

  if (start <= 0) {
    if (wholeLine) {
      if (x)
	*x = 0;
      if (y) {
	*y = firstLine->GetLocation();
	if (!top)
	  *y += firstLine->h;
      }
      return;
    }
    line = firstLine;
  } else if (start >= len) {
    if (extraLine && !eol) {
      if (y)
	*y = totalHeight - (top ? extraLineH : 0);
      if (x)
	*x = 0;
      return;
    } 

    line = lastLine;

    if (wholeLine) {
      if (x)
	*x = line->w;
      if (y) {
	*y = lastLine->GetLocation();
	if (!top)
	  *y += lastLine->h;
      }
      return;
    }
  } else {
    line = lineRoot->FindLine(PositionLine(start, eol));

    if (wholeLine) {
      if (y) {
	*y = line->GetLocation();
	if (!top)
	  *y += line->h;
      }
      if (!x)
	return;
    }
  }

  dc = NULL;

  topy = line->GetLocation();

  start -= line->GetPosition();
  if (!start) {
    horiz = 0;
    snip = line->snip;
  } else if (start == line->len) {
    horiz = line->w - line->lastW;
    snip = line->lastSnip;
  } else {
    /* linear seach for snip */
    snip = NULL;
    horiz = 0;
    
    while(1) {
      snip = snip ? snip->next : line->snip;
      
      if (start >= snip->count) {
	start -= snip->count;
	if (!dc) {
	  dc = admin->GetDC();
	  if (!dc)
	    return;
	}
	horiz += snip->PartialOffset(dc, horiz, topy, snip->count);
      } else
	break;
    }
  }


  if (x) {
    if (start && !dc) {
      dc = admin->GetDC();
      if (!dc)
	return;
    }
      
    *x = horiz + (start ? snip->PartialOffset(dc, horiz, topy, start) : 0);
  }

  if (!wholeLine && y) {
    if (!dc) {
      dc = admin->GetDC();
      if (!dc)
	return;
    }
    snip->GetExtent(dc, horiz, topy, NULL, &h, &descent, &space);
    align = snip->style->GetAlignment();
    if (align == wxALIGN_BOTTOM)
      *y = topy + line->bottombase + descent - (top ? h : 0);
    else if (align == wxALIGN_TOP)
      *y = topy + line->topbase - space + (top ? 0 : h);
    else {
      h = (h - descent - space) / 2;
      *y = topy + ((line->topbase + line->bottombase) / 2)
	+ (top ? (- h - space) : (h + descent));
    }
  }
}

float wxMediaEdit::LineLocation(long i, Bool top)
{
  wxMediaLine *line;
  float y;

  if (!CheckRecalc())
    return 0.0;

  if (i < 0)
    return 0;
  else if (i > numValidLines)
    return totalHeight;
  else if (i == numValidLines) {
    if (extraLine)
      return totalHeight - extraLineH;
    return totalHeight;
  }

  line = lineRoot->FindLine(i);
  y = line->GetLocation();

  if (top)
    return y;
  else
    return y + line->h;
}

long wxMediaEdit::LineStartPosition(long i, Bool visibleOnly)
{
  if (!CheckRecalc(maxWidth > 0))
    return FALSE;

  if (i < 0)
    i = 0;
  else if (i >= numValidLines) {
    if (extraLine)
      return len;
    i = numValidLines - 1;
  }

  if (visibleOnly)
    return FindFirstVisiblePosition(lineRoot->FindLine(i));
  else
    return lineRoot->FindLine(i)->GetPosition();
}

long wxMediaEdit::LineEndPosition(long i, Bool visibleOnly)
{
  wxMediaLine *line;
  long p;

  if (!CheckRecalc(maxWidth > 0))
    return FALSE;

  if (i < 0)
    i = 0;
  else if (i >= numValidLines) {
    if (extraLine)
      return len;
    i = numValidLines - 1;
  }

  line = lineRoot->FindLine(i);

  p = line->GetPosition() + line->len;
  if (visibleOnly)
    FindLastVisiblePosition(line, &p);

  return p;
}

long wxMediaEdit::LineLength(long i)
{
  if (i < 0)
    return 0;
  else if (i >= numValidLines)
    return 0;
  
  return lineRoot->FindLine(i)->len;
}

long wxMediaEdit::LastPosition(void)
{
  return len;
}

long wxMediaEdit::LastLine(void)
{
  return numValidLines - (extraLine ? 0 : 1);
}

void wxMediaEdit::GetExtent(float *w, float *h)
{
  CheckRecalc();

  if (w)
    *w = totalWidth;
  if (h)
    *h = totalHeight;
}

float wxMediaEdit::GetDescent(void)
{
  CheckRecalc();

  return finalDescent;
}

float wxMediaEdit::GetSpace(void)
{
  CheckRecalc();

  return initialSpace;
}

float wxMediaEdit::ScrollLineLocation(long scroll, float displayHeight)
{
  wxMediaLine *line;
  long p;
  float y;

  line = lineRoot->FindScroll(scroll);
  p = line->GetScroll();
  y = line->GetLocation();

  if (p < scroll)
    y += line->ScrollOffset(scroll - p);

  return y;
}

long wxMediaEdit::NumScrollLines(float displayHeight)
{
  return lastLine->GetScroll() + lastLine->numscrolls + (extraLine ? 1 : 0);
}

long wxMediaEdit::FindScrollLine(float p, float displayHeight)
{
  wxMediaLine *line;
  float y;
  long s;

  line = lineRoot->FindLocation(p);
  s = line->GetScroll();

  if (line->numscrolls > 1) {
    y = line->GetLocation();
    s += line->FindExtraScroll(p - y);
  }

  return s;
}

/****************************************************************/

long wxMediaEdit::FindString(char *str, int direction, long start, long end,
			     Bool bos, Bool caseSens)
{
  return _FindStringAll(str, direction, start, end, NULL, TRUE, bos, caseSens);
}

long *wxMediaEdit::FindStringAll(char *str, long *cnt, int direction,
				long start, long end, Bool bos, Bool caseSens)
{
  long *positions;

  *cnt = _FindStringAll(str, direction, start, end, &positions, FALSE,
			bos, caseSens);
  if (*cnt < 0) {
    *cnt = 0;
    positions = NULL;
  }
  return positions;
}

long wxMediaEdit::FindNewline(int direction, long start, long end)
{
  return FindString("\n", direction, start, end);
}

/****************************************************************/

void wxMediaEdit::ChangeStyle(wxStyleDelta *delta)
{
  ChangeStyle(delta, -1);
}

void wxMediaEdit::ChangeStyle(wxStyleDelta *delta, long start, long end)
{
  _ChangeStyle(start > -1 ? start : startpos, 
	       end > -1 ? end : (start > -1 ? start : endpos), 
	       NULL, delta);
}

void wxMediaEdit::ChangeStyle(wxStyle *style, long start, long end)
{
  _ChangeStyle(start > -1 ? start : startpos, 
	       end > -1 ? end : (start > -1 ? start : endpos), 
	       style, NULL);
}

void wxMediaEdit::SetStyleList(wxStyleList *newList)
{
  wxSnip *snip;
  wxStyle *style, *baseStyle, *newStyle;
  int count, index, baseIndex;
  wxStyle **map;
  wxStyleDelta delta;
  char *name;

  count = styleList->Number();
  map = new wxStyle*[count];
  for (index = 0; index < count; index++) {
    style = styleList->IndexToStyle(index);
    baseStyle = style->GetBaseStyle();
    baseIndex = styleList->StyleToIndex(baseStyle);

    style->GetDelta(delta);
    
    newStyle = newList->FindOrCreateStyle(map[baseIndex], &delta);
    if (name = style->GetName())
      newStyle = newList->NewNamedStyle(name, newStyle);

    map[index] = newStyle;
  }

  for (snip = snips; snip; snip = snip->next) {
    style = snip->style;
    index = styleList->StyleToIndex(style);
    snip->style = map[index];
  }

  wxMediaBuffer::SetStyleList(newList);

  SizeCacheInvalid();
  NeedRefresh(-1, -1);
}

void wxMediaEdit::StyleHasChanged(wxStyle *style)
{
  if (!style) {
    NeedRefresh(-1, -1); /* Our queue to repaint */
    return;
  }
  
  wxSnip *snip;
  
  for (snip = snips; snip; snip = snip->next) {
    if (PTREQ(snip->style, style)) {
      snip->style = style;
      snip->SizeCacheInvalid();
      snip->line->MarkRecalculate();
    }
  }
}

/****************************************************************/

Bool wxMediaEdit::ScrollTo(wxSnip *snip, float localx, float localy, 
			   float w, float h, Bool refresh, int bias)
{
  float x, y;

  if (delayRefresh) {
    delayedscroll = -1;
    delayedscrollbox = TRUE;
    delayedscrollsnip = snip;
    delayedscrollX = x;
    delayedscrollY = y;
    delayedscrollW = w;
    delayedscrollH = h;
    delayedscrollbias = bias;
    return FALSE;
  } else {
    if (!GetSnipPositionAndLocation(snip, NULL, &x, &y))
      return FALSE;
    if (admin->ScrollTo(x + localx, y + localy, w, h, refresh, bias)) {
      if (!refresh)
	refreshAll = TRUE;
      return TRUE;
    } else
      return FALSE;
  }
}

void wxMediaEdit::Resized(wxSnip *snip, Bool redraw_now)
{
  if (!GetSnipPositionAndLocation(snip, NULL, NULL, NULL))
    return;

  snip->line->MarkRecalculate();
  if (!graphicMaybeInvalid)
    graphicMaybeInvalid = TRUE;

  if (!redraw_now)
    delayRefresh++;
  RefreshByLineDemand();
  if (!redraw_now)
    --delayRefresh;
}

void wxMediaEdit::SetCaretOwner(wxSnip *snip)
{
  if (DoSetCaretOwner(snip))
    NeedCaretRefresh();
}

Bool wxMediaEdit::ReleaseSnip(wxSnip *snip)
{
  long pos;

  if ((pos = GetSnipPosition(snip)) < 0)
    return FALSE;

  _Delete(pos, pos + snip->count, FALSE);

  return TRUE;
}

void wxMediaEdit::RefreshBox(float L, float T, float w, float h)
{

  float B, R;

  B = T + h;
  R = L + w;

  if (refreshBoxUnset) {
    refreshL = L;
    refreshR = R;
    refreshT = T;
    refreshB = B;
    refreshBoxUnset = FALSE;
  } else {
    if (L < refreshL)
      refreshL = L;
    if (R > refreshR)
      refreshR = R;
    if (T < refreshT)
      refreshT = T;
    if (B > refreshB)
      refreshB = B;
  }

  drawCachedInBitmap = FALSE;
}

void wxMediaEdit::NeedsUpdate(wxSnip *snip, float localx, float localy, 
			      float w, float h)
{
  float x, y;

  if (!GetSnipLocation(snip, &x, &y))
    return;

  RefreshBox(x + localx, y + localy, w, h);

  if (!delayRefresh)
    Redraw();
}

void wxMediaEdit::InvalidateBitmapCache(float x, float y, float w, float h)
{
  if (w < 0)
    w = totalWidth - x;
  if (h < 0)
    h = totalHeight - y;

  RefreshBox(x, y, w, h);
  if (!delayRefresh)
    Redraw();
}
