/*
 * File:        wx_mpriv.cc
 * Purpose:     wxMediaEdit private methods implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_media.h"
#include "wx_timer.h"

#include <string.h>
#include <stdlib.h>
#include <fstream.h>
#include <ctype.h>

#include "wx_mpriv.h"

/* Correction when not using standard graphics: */
#ifdef wx_x
#if !WX_STANDARD_GRAPHICS
#define DELTA (-1)
#endif
#endif
#ifndef DELTA
#define DELTA 0
#endif

/* Debugging: */
#define CHECK_CONSISTENCY 0
#define LOOK_FOR_ZEROED 0
#define PLOT_SNIP_DOT 0

#ifndef ROUND
#include <math.h>
#define ROUND(x) (int)floor(x)
#endif

#define A_VERY_BIG_NUMBER 1e50

static wxCursor *iBeam = NULL, *arrow = NULL;

/****************************************************************/
/*                             PRIVATE                          */
/****************************************************************/

void wxMediaEdit::_SetPosition(Bool setflash, int bias, long start, long end, 
			       Bool ateol, Bool scroll)
{
  long oldstart, oldend, sPos;
  Bool oldateol;
  wxSnip *snip;
  Bool needRefresh;

  if (locked)
    return;

  EndStreaks(wxSTREAK_EXCEPT_DELAYED);
  
  if ((start < 0) 
      || ((end != -1) && (start > end)))
    return;

  if (end == -1)
    end = start;
  else if (end > len)
    end = len;

  if (start > len)
    start = len;

  if (ateol) {
    if (start != end)
      ateol = FALSE;
    else {
      snip = FindSnip(start, -1, &sPos);
      if (!(snip->flags & wxSNIP_NEWLINE) 
	  || (snip->flags & wxSNIP_INVISIBLE) 
	  || start != sPos + snip->count)
	ateol = FALSE;
    }
  }

  if (flash) {
    oldstart = flashstartpos;
    oldend = flashendpos;
    oldateol = flashposateol;
  } else {
    oldstart = startpos;
    oldend = endpos;
    oldateol = posateol;
  }
  
  if (!setflash && flash && flashautoreset) {
    flash = FALSE;
    if (flashTimer) {
      flashTimer->Stop();
      delete flashTimer;
      flashTimer = NULL;
    }
  }

  if (start == oldstart && end == oldend && ateol == oldateol)
    needRefresh = FALSE;
  else {
    needRefresh = TRUE;

    CheckMergeSnips(startpos);
    CheckMergeSnips(endpos);
    
    if (setflash) {
      flashstartpos = start;
      flashendpos = end;
      flashposateol = ateol;
    } else {
      caretStyle = NULL;
      startpos = start;
      endpos = end;
      posateol = ateol;
    }
  }

  if (setflash)
    flash = TRUE;

  if (scroll) {
    long scrollStart, scrollEnd;

    if (bias < -1) {
      scrollStart = scrollEnd = start;
      bias = 0;
    } else if (bias > 1) {
      scrollStart = scrollEnd = end;
      bias = 0;
    } else {
      scrollStart = start;
      scrollEnd = end;
    }
  
    if (ScrollToPosition(scrollStart, posateol, TRUE, scrollEnd, bias))
      needRefresh = FALSE;
  }

  if (needRefresh) {
    if (admin && (admin->standard > 0) && !delayRefresh 
	&& oldstart == oldend && start == end && caretOn
	&& caretLocationX >= 0 && !flash) {
      /* Try to take a shortcut */
      if (CaretOff()) {
	/* Shortcut works */
	caretLocationX = -1;
	CaretOn();
	needRefresh = FALSE;
      }
    }

    if (needRefresh) {
      if (start >= oldend || end <= oldstart) {
	/* No overlap: */
	NeedRefresh(oldstart, oldend);
	NeedRefresh(start, end);
      } else {
	if (start < oldstart)
	  NeedRefresh(start, oldstart);
	if (oldstart < start)
	  NeedRefresh(oldstart, start);
	if (end < oldend)
	  NeedRefresh(end, oldend);
	if (oldend < end)
	  NeedRefresh(oldend, end);
      }
    }
  }
}

void wxMediaFlashTimer::Notify(void)
{
  media->FlashOff();
}

/**********************************************************************/

void wxMediaEdit::_ChangeStyle(long start, long end, 
			       wxStyle *newStyle, wxStyleDelta *delta)
{
  wxSnip *gsnip, *startSnip, *endSnip;
  wxStyleChangeRecord *rec;
  wxStyle *style, *style2;
  long p;
  int something;

  if (locked || userLocked)
    return;
  
  if (newStyle && (styleList->StyleToIndex(newStyle) < 0))
    return;

  if (start < 0)
    start = 0;
  if (start > len)
    start = len;
  if (end > len)
    end = len;
  if (start > end)
    return;

  if (!newStyle && !delta)
    newStyle = styleList->FindNamedStyle(STD_STYLE);

  if ((startpos == start) && (endpos == end) && (start == end)) {
    if (newStyle)
      caretStyle = newStyle;
    else {
      gsnip = FindSnip(start, -1);
      caretStyle = styleList->FindOrCreateStyle(gsnip->style, delta);
    }
    return;
  }

  MakeSnipset(start, end);
  
  startSnip = FindSnip(start, +1);
  endSnip = FindSnip(end, +2);

  rec = new wxStyleChangeRecord(start, end, delayedStreak || !modified);

  something = FALSE;

  p = start;
  for (gsnip = startSnip; PTRNE(gsnip, endSnip); gsnip = gsnip->next) {
    style = gsnip->style;

    if (newStyle)
      style2 = newStyle;
    else
      style2 = styleList->FindOrCreateStyle(style, delta);

    if (PTRNE(style, style2)) {
      gsnip->style = style2;
      rec->AddStyleChange(p, p + gsnip->count, style);

      gsnip->SizeCacheInvalid();
      gsnip->line->MarkRecalculate();
      if (maxWidth > 0)
	gsnip->line->MarkCheckFlow();

      something = TRUE;
    }

    p += gsnip->count;
  }

  if (something) {
    if (startSnip->line->prev
	&& !(startSnip->line->prev->lastSnip->flags & wxSNIP_HARD_NEWLINE))
      startSnip->line->prev->MarkCheckFlow();
    
    if (!modified)
      AddUndo(new wxUnmodifyRecord);
    AddUndo(rec);
    if (delayRefresh)
      delayedStreak = TRUE;
    
    CheckMergeSnips(start);
    CheckMergeSnips(end);
    
    modified = TRUE;

    RefreshByLineDemand();
  } else
    delete rec;
}

/****************************************************************/

static void MakeCursors(void)
{
  if (!iBeam) {
    arrow = new wxCursor(wxCURSOR_ARROW);
#ifdef wx_xview
    iBeam = arrow; /* Because the ibeam looks bad, and internal scrolls
		      don't work well */
#else
    iBeam = new wxCursor(wxCURSOR_IBEAM);
#endif
  }
}

void wxMediaEdit::IBeam(Bool on)
{
  if (!admin)
    return;

  MakeCursors();

  if (on != ibeamCursor) {
    if (on)
      admin->SetCursor(iBeam);
    else
      admin->SetCursor(arrow);
    ibeamCursor = on;
  }
}

void wxMediaEdit::SettingAdmin(wxMediaAdmin *newadmin)
{
  caretLocationX = -1;

  if (admin && admin->standard)
    IBeam(FALSE);

  if (newadmin && newadmin->standard) {
    MakeCursors();
    newadmin->SetCursor(iBeam);
    ibeamCursor = TRUE;
  }
}

void wxMediaEdit::EndStreaks(int exception)
{
  if (exception != wxSTREAK_KEY_SEQUENCE)
    map->BreakSequence();
  if (flash && (flashautoreset > 0))
    FlashOff();

  typingStreak = FALSE;
  deletionStreak = FALSE;
  vcursorStreak = FALSE;  
  anchorStreak = FALSE;  

  if (exception != wxSTREAK_EXCEPT_DELAYED)
    delayedStreak = FALSE;

  killStreak = FALSE;
}


/****************************************************************/

long wxMediaEdit::_FindPositionInLine(Bool internal, long i, float x, 
				      Bool *ateol, Bool *onit)
{
  float w, X, topy;
  wxSnip *snip;
  wxDC *dc;
  long p, sPos;
  wxMediaLine *line;
  Bool atsnipend;

  if (onit)
    *onit = FALSE;

  if (!internal && !CheckRecalc())
    return 0;

  if (i < 0)
    return 0;
  if (i >= numValidLines)
    return len;

  line = lineRoot->FindLine(i);

  if (ateol)
    *ateol = FALSE;

  if (x <= 0)
    return FindFirstVisiblePosition(line);

  p = line->GetPosition();

  if (x >= line->w) {
    /* snip == the last one */
    snip = line->lastSnip;
    sPos = p + line->len - snip->count;
    p += line->len;
  } else {
    if (onit)
      *onit = TRUE;

    dc = admin->GetDC();
    if (!dc)
      return 0;
    
    X = 0;

    /* linear seach for snip */
    snip = NULL;
    topy = line->GetLocation();
    while(1) {
      snip = snip ? snip->next : line->snip;

      snip->GetExtent(dc, X, topy, &w);

      if (x > w && snip->next) {
	x -= w;
	X += w;
	p += snip->count;
      } else {
	/* Found the right snip */
	sPos = p;
	p += _FindPositionInSnip(dc, X, topy, snip, x);
	break;
      }
    }
  }

  /* Back up over invisibles */
  atsnipend = (p - sPos == snip->count);
  if (atsnipend)
    FindLastVisiblePosition(line, &p, &snip);

  if (ateol && atsnipend && snip && PTREQ(snip, line->lastSnip))
    *ateol = TRUE;

  return p;
}

long wxMediaEdit::FindFirstVisiblePosition(wxMediaLine *line, wxSnip *snip)
{
  wxSnip *nextSnip;
  long p, startp;

  if (!snip)
    snip = line->snip;

  p = startp = line->GetPosition();
  nextSnip = line->lastSnip->next;

  while (PTRNE(snip, nextSnip)) {
    if (snip->flags & wxSNIP_INVISIBLE)
      p += snip->count;
    else
      break;
    snip = snip->next;
  }

  if (PTREQ(snip, nextSnip)) {
    /* If everything is invisible, then presumably the CR is forced,
       so go to the beginning of the line anyway */
    p = startp;
  }

  return p;
}

void wxMediaEdit::FindLastVisiblePosition(wxMediaLine *line, long *p, 
					  wxSnip **snipP)
{
  wxSnip *snip;
  
  snip = snipP ? *snipP : NULL;

  if (!snip)
    snip = line->lastSnip;

  do {
    if (snip->flags & wxSNIP_INVISIBLE) {
      *p -= snip->count;
      if (PTRNE(snip, line->snip))
	snip = snip->prev;
    }
  } while ((snip->flags & wxSNIP_INVISIBLE) && PTRNE(snip, line->snip));

  if (snipP)
    *snipP = snip;
}

/****************************************************************/

long wxMediaEdit::_FindStringAll(char *str, int direction, 
				 long start, long end,
				 long **positions, Bool justOne,
				 Bool bos, Bool caseSens)
{
  wxSnip *snip;
  char text[256], *oldStr, c;
  long *smap;
  long sPos, p, n, thistime, thisoffset, need, checked, offset, shorten, i;
  long slen, s, sbase, beyond, sgoal, totalCount, allocFound, foundCount;

  if (!direction)
    direction = 1;
  if (direction < -1)
    direction = -1;
  else if (direction > 1)
    direction = 1;

  if (start < 0)
    start = startpos;
  if (end < 0) {
    if (direction < 0)
      end = 0;
    else
      end = len;
  }
  if (start > len)
    start = len;
  if (end > len)
    end = len;

  if (direction < 0)
    totalCount = start - end;
  else
    totalCount = end - start;
  if (totalCount < 0)
    return -1;

  slen = strlen(str);
  if (!slen)
    return -1;

  if (!caseSens) {
    oldStr = str;
    str = new char[slen + 1];
    for (i = 0; i < slen; i++)
      str[i] = tolower(oldStr[i]);
    str[i] = 0;
  }
  
  snip = FindSnip(start, direction, &sPos);
  if (!snip)
    return -1;

  if (direction > 0) {
    offset = start - sPos;
    shorten = 0;
    sbase = 0;
    beyond = -1;
    sgoal = slen;
  } else {
    shorten = (sPos + snip->count) - start;
    offset = 0;
    sbase = slen - 1;
    beyond = slen;
    sgoal = -1;
  }

  smap = new long[slen];

  smap[sbase] = beyond;
  s = beyond;
  for (i = sbase + direction; i != sgoal; i += direction) {
    while ((s != beyond) && (str[s + direction] != str[i]))
      s = smap[s];
    if (str[s + direction] == str[i])
      s += direction;
    smap[i] = s;
  }

  s = beyond;

  if (!justOne) {
    allocFound = 10;
    *positions = new long[allocFound];
    foundCount = 0;
  }

  while (snip && totalCount) {
    need = snip->count - shorten - offset;
    if (need > totalCount) {
      if (direction < 0)
	offset += (need - totalCount);
      need = totalCount;
    }

    checked = 0;

    totalCount -= need;

    do {
      thistime = need;
      if (thistime > 255)
	thistime = 255;
      need -= thistime;

      thisoffset = offset + ((direction < 0) ? need : checked);

      snip->GetText((char *)text, thisoffset, thistime);
      text[thistime] = 0;
      
      i = (direction > 0) ? 0 : thistime - 1;
      n = thistime;
      while(n--) {
	c = text[i];
	if (!caseSens)
	  if (c >= 'A' && c <= 'Z')
	    c += ('a' - 'A');
	while ((s != beyond) && (str[s + direction] != c))
	  s = smap[s];
	if (str[s + direction] == c) {
	  s += direction;
	  if (s + direction == sgoal) {
	    p = sPos + i + thisoffset;
	    if (bos) {
	      if (direction < 0)
		p += slen;
	      else
		p -= (slen - 1);
	    }
	    if (justOne)
	      goto search_done;
	    else {
	      if (foundCount == allocFound) {
		long *old = *positions, oldCount = allocFound;

		allocFound *= 2;
		*positions = new long[allocFound];

		memcpy(*positions, old, oldCount * sizeof(long));
	      }
	      (*positions)[foundCount++] = p;
	      s = beyond;
	    }
	  }
	}
	i += direction;
      }

      checked += thistime;
    } while(need);

    if (direction > 0)
      sPos += snip->count;
    snip = (direction > 0) ? snip->next : snip->prev;
    if (snip) {
      if (direction < 0)
	sPos -= snip->count;
    }
    offset = shorten = 0;
  }

  p = -1;

 search_done:
  delete[] smap;

  return justOne ? p : foundCount;
}

/****************************************************************/

void wxMediaEdit::MakeOnlySnip(void)
{
  snips = new wxTextSnip();
  snips->style = styleList->FindNamedStyle(STD_STYLE);
  snips->count = 0;
  snips->SetAdmin(snipAdmin);
  snips->prev = NULL;
  snips->next = NULL;

  snips->line = lineRoot = firstLine = lastLine = new wxMediaLine;

  lineRoot->snip = lineRoot->lastSnip = snips;

  lastSnip = snips;
  snipCount = 1;

  numValidLines = 1;
}

void wxMediaEdit::SpliceSnip(wxSnip *snip, wxSnip *prev, wxSnip *next)
{
  if (prev)
    prev->next = snip;
  else
    snips = snip;
  snip->prev = prev;
  snip->next = next;
  if (next)
    next->prev = snip;
  else
    lastSnip = snip;
  
}

void wxMediaEdit::InsertSnip(wxSnip *before, wxSnip *snip)
{
  if (PTREQ(snips, lastSnip) && !snips->count) {
    AppendSnip(snip);
  } else {
    SpliceSnip(snip, before->prev, before);
    snipCount++;
  }
}

void wxMediaEdit::AppendSnip(wxSnip *snip)
{
  if (PTREQ(snips, lastSnip) && !snips->count) {
    /* Get rid of empty snip */
    delete snips;
    snips = lastSnip = snip;
  } else {
    SpliceSnip(snip, lastSnip, NULL);
    snipCount++;
  }
}

void wxMediaEdit::DeleteSnip(wxSnip *snip)
{
  if (snip->next)
    SpliceSnip(snip->next, snip->prev, snip->next->next);
  else if (snip->prev)
    SpliceSnip(snip->prev, snip->prev->prev, snip->next);
  else {
    lastSnip = snips = NULL;
  }
  --snipCount;
  snip->flags -= wxSNIP_OWNED;
  snip->SetAdmin(NULL);
  snip->flags += wxSNIP_OWNED;
}

/****************************************************************/

wxSnip *wxMediaEdit::FindSnip(long p, int direction, long *sPos)
{
  wxSnip *snip;
  wxMediaLine *line;
  long pos;

  if ((direction < -1) && !p)
    return NULL;

  line = lineRoot->FindPosition(p);
  pos = line->GetPosition();
  p -= pos;
  if (sPos)
    *sPos = pos;

  snip = line->snip;
  if (!p && snip->prev) {
    /* Back up one: */
    snip = snip->prev;
    p += snip->count;
    if (sPos)
      *sPos -= snip->count;
  }

  for (; snip; snip = snip->next) {
    p -= snip->count;
    if ((!direction && !p)
	|| ((direction < 0) && (p <= 0))
	|| ((direction > 0) && (p < 0)))
      return snip;

    if (!direction && (p < 0))
      return NULL;

    if (sPos)
      *sPos += snip->count;
  }

  if (direction < 2)
    return lastSnip;
  else
    return NULL;
}

void wxMediaEdit::MakeSnipset(long start, long end) 
{
  long sPos, c;
  wxSnip *prev, *next, *snip, *insSnip;
  wxStyle *style;
  wxMediaLine *line;
  Bool atStart, atEnd;

  snip = FindSnip(start, +1, &sPos);
  if (start != sPos) {
    c = snip->count;
    line = snip->line;
    prev = snip->prev; next = snip->next;
    style = snip->style;

    atStart = PTREQ(line->snip, snip);
    atEnd = PTREQ(line->lastSnip, snip);

    snip->Split(start - sPos, &insSnip, &snip);

    /* Make *sure* that count is right */
    insSnip->count = start - sPos;
    snip->count = c - insSnip->count;

    snip->style = insSnip->style = style;

    snip->line = insSnip->line = line;
    if (atStart)
      line->snip = insSnip;
    if (atEnd)
      line->lastSnip = snip;

    SpliceSnip(snip, prev, next);
    InsertSnip(snip, insSnip);

    snip->SetAdmin(snipAdmin);
    insSnip->SetAdmin(snipAdmin);
  }

  snip = FindSnip(end, -1, &sPos);
  if (end != sPos + snip->count) {
    c = snip->count;
    line = snip->line;
    prev = snip->prev; next = snip->next;
    style = snip->style;

    atStart = PTREQ(line->snip, snip);
    atEnd = PTREQ(line->lastSnip, snip);

    snip->Split(end - sPos, &insSnip, &snip);

    /* Make *sure* that count is right */
    insSnip->count = end - sPos;
    snip->count = c - insSnip->count;

    snip->style = insSnip->style = style;

    snip->line = insSnip->line = line;
    if (atStart)
      line->snip = insSnip;
    if (atEnd)
      line->lastSnip = snip;

    SpliceSnip(snip, prev, next);
    InsertSnip(snip, insSnip);

    snip->SetAdmin(snipAdmin);
    insSnip->SetAdmin(snipAdmin);
  }
}

wxTextSnip *wxMediaEdit::InsertTextSnip(long start, wxStyle *style)
{
  long sPos, c;
  wxSnip *gsnip, *insGsnip, *prev, *next;
  wxTextSnip *snip;
  wxStyle *gstyle;
  wxMediaLine *line;
  Bool atStart, atEnd;

  snip = new wxTextSnip();
  snip->style = (style ? style : styleList->FindNamedStyle(STD_STYLE));
  snip->SetAdmin(snipAdmin);
  snip->count = 0;

  gsnip = FindSnip(start, +2, &sPos);
  if (!gsnip) {
    AppendSnip(snip);
    snip->line = lastLine;
    if (PTREQ(lastLine->snip, lastSnip))
      lastLine->snip = lastLine->lastSnip = snip;
    else
      lastLine->lastSnip = snip;
  } else if (PTREQ(start, sPos)) {
    InsertSnip(gsnip, snip);
    snip->line = gsnip->line;
    if (PTREQ(snip->line->snip, gsnip))
      snip->line->snip = snip;
  } else {
    c = gsnip->count;
    prev = gsnip->prev; next = gsnip->next;
    gstyle = gsnip->style;
    line = gsnip->line;

    atStart = PTREQ(line->snip, gsnip);
    atEnd = PTREQ(line->lastSnip, gsnip);

    
    gsnip->Split(start - sPos, &insGsnip, &gsnip);

    /* Make *sure* that count is right */
    insGsnip->count = start - sPos;
    gsnip->count = c - insGsnip->count;
    
    gsnip->style = insGsnip->style = gstyle;

    gsnip->line = insGsnip->line = snip->line = line;
    if (atStart)
      line->snip = insGsnip;
    if (atEnd)
      line->lastSnip = gsnip;

    SpliceSnip(gsnip, prev, next);
    InsertSnip(gsnip, snip);
    InsertSnip(snip, insGsnip);

    gsnip->SetAdmin(snipAdmin);
    insGsnip->SetAdmin(snipAdmin);
  }

  return snip;
}

void wxMediaEdit::CheckMergeSnips(long start)
{
  wxSnip *snip1, *snip2, *prev, *next;
  long sPos1, sPos2, c;

  wxMediaLine *line;
  Bool atStart, atEnd;

  snip1 = FindSnip(start, -1, &sPos1);
  snip2 = FindSnip(start, +1, &sPos2);

  if (PTRNE(snip1, snip2)) {
    if (snip1->snipclass && PTREQ(snip1->snipclass, snip2->snipclass)
	&& PTREQ(snip1->style, snip2->style)) {
      if (!(snip1->flags & wxSNIP_NEWLINE)
	  && (snip1->flags & wxSNIP_CAN_APPEND) 
	  && (snip2->flags & wxSNIP_CAN_APPEND)) {
	if (!snip1->count) {
	  if (PTREQ(snip1->line->snip, snip1))
	    snip1->line->snip = snip2;
	  DeleteSnip(snip1);
	  CheckMergeSnips(start);
	} else if (!snip2->count) {
	  if (PTREQ(snip1->line->lastSnip, snip2)) {
	    snip1->line->lastSnip = snip1;
	    snip1->line->MarkRecalculate(); // need lastW updated
	    graphicMaybeInvalid = TRUE;
	  }
	  DeleteSnip(snip2);
	  CheckMergeSnips(start);
	} else {
	  c = snip1->count + snip2->count;
	  prev = snip1->prev;
	  next = snip2->next;
	  line = snip1->line;
	  atStart = PTREQ(line->snip, snip1);
	  atEnd = PTREQ(line->lastSnip, snip2);
	  if (snip1 = snip2->MergeWith(snip1)) {
	    SpliceSnip(snip1, prev, next);
	    --snipCount;
	    /* Make *sure* that count is right */
	    snip1->count = c;

	    snip1->line = line;
	    if (atStart)
	      line->snip = snip1;
	    if (atEnd) {
	      line->lastSnip = snip1;
	      line->MarkRecalculate(); // need lastW updated
	      graphicMaybeInvalid = TRUE;
	    }
	  }
	}
      }
    }
  }
}

/****************************************************************/

Bool wxMediaEdit::GetSnipPositionAndLocation(wxSnip *thesnip, long *pos, 
					     float *x, float *y)
{
  wxSnip *snip;
  long p;

  if (!CheckRecalc(x || y))
    return FALSE;

  if (!thesnip->line || PTRNE(thesnip->line->GetRoot(), lineRoot))
    return FALSE;

  p = thesnip->line->GetPosition();
    
  for (snip = thesnip->line->snip; PTRNE(snip, thesnip); snip = snip->next)
    p += snip->count;

  if (pos)
    *pos = p;

  if (x || y) 
    PositionLocation(p, x, y);

  return TRUE;
}

Bool wxMediaEdit::GetSnipLocation(wxSnip *thesnip, float *x, float *y)
{
  return GetSnipPositionAndLocation(thesnip, NULL, x, y);
}

long wxMediaEdit::GetSnipPosition(wxSnip *thesnip)
{
  long pos;

  if (GetSnipPositionAndLocation(thesnip, &pos))
    return pos;
  else
    return -1;
}

/****************************************************************/

void wxMediaEdit::AdjustClickbacks(long start, long end, 
				   long d, wxDeleteRecord *rec)
{
  wxNode *node, *next;
  wxClickback *click;
  Bool deleteit;

  for (node = clickbacks->First(); node; node = next) {
    next = node->Next();
    click = (wxClickback *)node->Data();
    deleteit = FALSE;
    if (click->start >= start && click->end <= end) {
      deleteit = TRUE;
    } else if (click->start >= end) {
      click->start += d;
      click->end += d;
    } else if (click->start <= start && click->end > end)
      click->end += d;
    else if (click->start > start && click->end > end) {
      click->start = start;
      click->end += d;
    }

    if (click->end == click->start)
      deleteit = TRUE;
    if (deleteit) {
      clickbacks->DeleteNode(node);
      if (rec)
	rec->AddClickback(click);
      else
	delete click;
    }
  }
}

wxClickback *wxMediaEdit::FindClickback(long start)
{
  wxNode *node;
  wxClickback *click;

  for (node = clickbacks->First(); node; node = node->Next()) {
    click = (wxClickback *)node->Data();
    if (click->start <= start && click->end > start)
      return click;
  }

  return NULL;
}

void wxMediaEdit::SetClickbackHilited(wxClickback *click, Bool on)
{
  if (on != click->hilited) {
    if (on) {
      interceptmode = TRUE;
      intercepted = new wxList();
      
      BeginEditSequence();
      FlashOn(click->start, click->end, FALSE, FALSE, -1);
      ChangeStyle(click->delta, click->start, click->end);
      EndEditSequence();

      click->unhilite = intercepted;
      interceptmode = FALSE;
    } else {
      PerformUndoList(click->unhilite);

      wxNode *node;
      for (node = click->unhilite->First(); node; node = node->Next())
	delete (wxChangeRecord *)node->Data();
  
      delete click->unhilite;
      FlashOff();
    }
    click->hilited = on;
  }
}

/****************************************************************/

Bool wxMediaEdit::CheckRecalc(Bool need_graphic)
{
  if (locked)
    return FALSE;

  if (need_graphic) {
    if (!admin)
      return FALSE;

    if (graphicMaybeInvalid) {
      wxDC *dc;

      dc = admin->GetDC();
      if (!dc)
	return FALSE;
      
      RecalcLines(dc, need_graphic);
    }
  }

  return TRUE;
}

Bool wxMediaEdit::CheckFlow(float maxw, wxDC *dc, float Y, 
			    long startp, wxSnip *start)
{
  long p, c, origc, b;
  float totalWidth, w;
  wxSnip *snip;
  Bool checkingUnderflow; // no overflow => check move up from next line
  Bool checkingUnderflowAtNext;
  Bool noChangeIfEndOfSnip, noChangeIfStartOfSnip;
  Bool theFirstSnip, firstUnderflow, hadNewline;

  totalWidth = 0;
  p = startp;
  checkingUnderflow = FALSE; // start by insuring no overflow
  checkingUnderflowAtNext = FALSE;
  noChangeIfEndOfSnip = TRUE; // Because an immediate overflow can't be helped
  noChangeIfStartOfSnip = FALSE;
  theFirstSnip = TRUE;

  for (snip = start; 
       snip && !(snip->flags & wxSNIP_HARD_NEWLINE); 
       snip = snip->next) {

    if (!checkingUnderflow) {
      if (checkingUnderflow = checkingUnderflowAtNext)
	firstUnderflow = TRUE;
    }
    noChangeIfStartOfSnip = noChangeIfEndOfSnip;

    if (snip->flags & wxSNIP_NEWLINE) {
      noChangeIfEndOfSnip = !checkingUnderflow;
      snip->flags -= wxSNIP_NEWLINE;
      checkingUnderflowAtNext = TRUE;
      hadNewline = TRUE;
    } else {
      noChangeIfEndOfSnip = FALSE;
      checkingUnderflowAtNext = FALSE;
      hadNewline = FALSE;
    }

    snip->GetExtent(dc, totalWidth, Y, &w);
    totalWidth += w;
    if (totalWidth > maxw) {
      totalWidth -= w;
      /* Get best breaking position: */
      origc = _FindPositionInSnip(dc, totalWidth, Y, snip, maxw - totalWidth);
      
      /* get legal breaking position before optimal: */
      b = p + origc + 1;
      FindWordbreak(&b, NULL, wxBREAK_FOR_LINE);
      c = b - p;

      if (c <= 0) {
	if ((b <= startp) && checkingUnderflow && origc) {
	  /* The word was currently force-broken; shift some part to here */
	  p = p + origc;
	} else if ((checkingUnderflow && firstUnderflow
		    && (b <= startp || c>= 0))
		   || (!theFirstSnip && !c)) {
	  /* Can't fit this snip in the line */
	  if (snip->prev)
	    snip->prev->flags |= wxSNIP_NEWLINE;
	  if (hadNewline)
	    snip->flags |= wxSNIP_NEWLINE;
	  if (noChangeIfStartOfSnip)
	    return FALSE;
	  refreshAll = TRUE;
	  return TRUE;
	} else if ((c < 0) && (b > startp)) {
	  /* Overflow, but previous wordbreak was before this snip */
	  p = b;
	} else {
	  /* Overflow: we have to break the word anyway */
	  p = p + origc;
	}
      } else {
	p = p + c;
      }
      MakeSnipset(p, p);
      snip = FindSnip(p, -1);
      snip->flags |= wxSNIP_NEWLINE;
      refreshAll = TRUE;
      return TRUE;
    }
    
    p += snip->count;

    theFirstSnip = FALSE;
    firstUnderflow = FALSE;
  }

  if (!checkingUnderflow || noChangeIfEndOfSnip)
    return FALSE;
  else {
    refreshAll = TRUE;
    return TRUE;
  }
}

#if CHECK_CONSISTENCY
static int CheckRBConsistent(wxMediaLine *line)
{
  Bool red;
  int l, r;

  if (line == NIL)
    return 0;

  red = (line->flags & 1);

  if (red) {
    if (line->left->flags & 1)
      fprintf(stderr, "Red left inconsistency\n");
    if (line->right->flags & 1)
      fprintf(stderr, "Red right inconsistency\n");
  }

  l = CheckRBConsistent(line->left);
  r = CheckRBConsistent(line->right);

  if (l != r)
    fprintf(stderr, "Black count inconsistency\n");
  
  return l + (red ? 0 : 1);
}
#endif

void wxMediaEdit::RecalcLines(wxDC *dc, Bool calcGraphics)
{
  wxMediaLine *line;
  wxSnip *snip;
  float X, Y, descent, space;
  Bool changed, resized;

  if (!calcGraphics)
    return;

  changed = FALSE;

#if CHECK_CONSISTENCY
  long p, p2;

  snip = snips;

  if (firstLine != lineRoot->First())
    fprintf(stderr, "First line inconsistency\n");
  if (lastLine != lineRoot->Last())
    fprintf(stderr, "Last line inconsistency\n");

  if (firstLine->prev)
    fprintf(stderr, "First line prev inconsistency\n");
  if (lastLine->next)
    fprintf(stderr, "First line next inconsistency\n");

  if (firstLine->snip != snips)
    fprintf(stderr, "First line first snip inconsistency\n");
  if (lastLine->lastSnip != lastSnip)
    fprintf(stderr, "Last line last snip inconsistency\n");

  CheckRBConsistent(lineRoot);

  Y = 0;
  p = 0;
  for (line = firstLine; line; line = line->next) {
    if (line->snip != snip)
      fprintf(stderr, "Line start inconsistency\n");
    
    if (line->GetPosition() != p)
      fprintf(stderr, "Line position inconsistency\n");
    if (line->GetLocation() != Y)
      fprintf(stderr, "Line location inconsistency\n");

    if (line->prev && (line->snip != line->prev->lastSnip->next))
      fprintf(stderr, "Line snip continuity inconsistency\n");

    p2 = p;
    for (snip = line->snip; snip && snip != line->lastSnip; snip = snip->next){
      if (snip->line != line)
	fprintf(stderr, "Snip's line inconsistency\n");
      p2 += snip->count;
    }
    if (snip->line != line)
	fprintf(stderr, "Snip's line inconsistency\n");

    if (!snip)
      fprintf(stderr, "Line end inconsistency\n");

    p2 += snip->count;
    if ((p2 - p) != line->len)
      fprintf(stderr, "Line count inconsistency\n");

    snip = line->lastSnip->next;

    if (!line->next && line != lastLine)
      fprintf(stderr, "Last line link inconsistency\n");

    wxMediaLine *other;
    if (line->left != NIL) {
      for (other=line->prev; other && other != line->left; other=other->prev);
      if (!other)
	fprintf(stderr, "Left link inconsistency\n");
    }
    if (line->right != NIL) {
      for (other=line->next; other && other!=line->right; other=other->next);
      if (!other)
	fprintf(stderr, "Right link inconsistency\n");
    }
    if (line->parent == NIL) {
      if (line != lineRoot)
	fprintf(stderr, "Root inconsistency\n");
    } else if (line != line->parent->right && line != line->parent->left)
      fprintf(stderr, "Parent inconsistency\n");

    Y += line->h;
    p += line->len;
  }
  if (p != len)
    fprintf(stderr, "Total count inconsistency\n");

#endif

  if (snipCacheInvalid)
    for (snip = snips; snip; snip = snip->next)
      snip->SizeCacheInvalid();

  float old_max_width = maxWidth;

  if (flowInvalid && (maxWidth <= 0))
    maxWidth = A_VERY_BIG_NUMBER;

  if (graphicsInvalid || flowInvalid || snipCacheInvalid) {
    /* Set all lines invalid. */
    for (line = firstLine; line; line = line->next) {
      line->MarkRecalculate();
      if (flowInvalid)
	line->MarkCheckFlow();
    }
  }

  locked = TRUE;

#if LOOK_FOR_ZEROED
  if (len)
    for (snip = snips, i = 0; snip; snip = snip->next)
      if (!snip->count)
	fprintf(stderr, "zero snip found at %d\n", i);
      else
	i += snip->count;
#endif

  if (maxWidth > 0) {
    float w;
    w = maxWidth - CURSOR_WIDTH;
    while (lineRoot->UpdateFlow(&lineRoot, this, w, dc))
      changed = TRUE;
  }

  if (old_max_width != maxWidth)
    maxWidth = old_max_width;

  if (changed) {
    refreshAll = TRUE;
    firstLine = lineRoot->First();
    lastLine = lineRoot->Last();
    numValidLines = lineRoot->Number();
  }

  if (lineRoot->UpdateGraphics(this, dc))
    changed = TRUE;

  if (!changed && (graphicMaybeInvalid >= 0)) {
    graphicMaybeInvalid = FALSE;
    locked = FALSE;
    return;
  }

  graphicMaybeInvalid = FALSE;

  Y = lastLine->GetLocation() + lastLine->h;

  if (lastSnip->flags & wxSNIP_NEWLINE) {
    extraLine = TRUE;
    extraLineH = lastLine->lastH + lineSpacing;
    Y += extraLineH;
  } else {
    extraLine = FALSE;
    extraLineH = 0;
  }

  X = lineRoot->maxWidth + CURSOR_WIDTH;
  if (minWidth > 0 && X < minWidth)
    X = minWidth;

  if (minHeight > 0 && Y < minHeight)
    Y = minHeight;
  if (maxHeight > 0 && Y > maxHeight)
    Y = maxHeight;

  descent = lastLine->h - lastLine->bottombase;
  space = lastLine->topbase;
  
  if (totalHeight != Y || totalWidth != X 
      || finalDescent != descent || initialSpace != space) {
    totalHeight = Y;
    totalWidth = X;
    finalDescent = descent;
    initialSpace = space;
    resized = TRUE;
  } else
    resized = FALSE;

  graphicsInvalid = FALSE;  
  flowInvalid = FALSE;
  snipCacheInvalid = FALSE;

  drawCachedInBitmap = FALSE;

  locked = FALSE;

  if (resized && admin)
    admin->Resized(FALSE);
}

static skipBox = FALSE;

static wxPen *caretPen = NULL;

/* This does the actual drawing */
void wxMediaEdit::Redraw(wxDC *dc, float starty, float endy, 
			 float leftx, float rightx,
			 float dy, float dx, Bool show_caret)
{
  wxMediaLine *line;
  wxSnip *snip, *first, *last;
  float x, topbase, bottombase, hxs, hxe, hsxs, hsxe, hsys, hsye, down, bottom;
  float tleftx, tstarty, trightx, tendy;
  float h, w, descent, space, ycounter;
  long p, pcounter;
  long startpos, endpos;
  Bool posateol;
  Bool hilite, hiliteSome;
  int align;
  wxStyle *oldStyle;
  wxPen *savePen;
  wxBrush *saveBrush;
  static wxPen *outlinePen = NULL;
  static wxBrush *outlineBrush = NULL;

  if (locked)
    return;

  locked = TRUE;

  if (flash) {
    startpos = wxMediaEdit::flashstartpos;
    endpos = wxMediaEdit::flashendpos;
    posateol = wxMediaEdit::flashposateol;
  } else {
    startpos = wxMediaEdit::startpos;
    endpos = wxMediaEdit::endpos;
    posateol = wxMediaEdit::posateol;
  }

  if (!outlinePen) {
    outlinePen = new wxPen("BLACK", 0, wxTRANSPARENT);
    caretPen = new wxPen("BLACK", 0, wxSOLID);
    outlineBrush = new wxBrush("BLACK", wxSOLID);
  }
  
  dc->SetBackgroundMode(wxSOLID);

  dc->BeginDrawing();

  line = lineRoot->FindLocation(starty);

  if (!skipBox) {
    dc->SetLogicalFunction(wxCLEAR);
    dc->DrawRectangle(leftx + dx, starty + dy, 
		      rightx - leftx + DELTA, endy - starty + DELTA);
    dc->SetLogicalFunction(wxCOPY);
  }

  OnPaint(TRUE, dc, leftx, starty, rightx, endy, dx, dy, show_caret);

  if (!line)
    goto paint_done;

  tleftx = leftx + dx;
  tstarty = starty + dy;
  trightx = rightx + dx;
  tendy = endy + dy;

  oldStyle = NULL;

  ycounter = line->GetLocation();
  pcounter = line->GetPosition();

  for (; line; line = line->next) {
    if (ycounter >= endy)
      goto paint_done;

    first = line->snip;
    last = line->lastSnip->next;

    x = 0;
    bottombase = ycounter + line->bottombase;
    topbase = ycounter + line->topbase;
    p = pcounter;

    hiliteSome = FALSE;

    for (snip = first; PTRNE(snip, last); snip = snip->next) {
      
      snip->GetExtent(dc, x, ycounter, &w, &h, &descent, &space);
      
      align = snip->style->GetAlignment();
      
      if (align == wxALIGN_BOTTOM)
	down = bottombase - h + descent;
      else if (align == wxALIGN_TOP)
	down = topbase - space;
      else {
	down = (h - descent - space) / 2;
	down = (topbase + bottombase) / 2 - down - space;
      }

      snip->style->SwitchTo(dc, oldStyle);
      oldStyle = snip->style;

      snip->Draw(dc, x + dx, down + dy, 
		 tleftx, tstarty, trightx, tendy, 
		 dx, dy, PTREQ(snip, caretSnip) && show_caret);

#if PLOT_SNIP_DOT
      dc->DrawLine(x + dx, down + dy, x + dx + 2, down + dy + 2);
#endif

      /* The rules for hiliting are surprisingly complicated: */

      if (hiliteOn && show_caret && !caretSnip) {
	if (posateol)
	  hilite = (startpos == p + snip->count);
	else
	  hilite = (((startpos < p + snip->count) && (endpos >= p)
		     && (startpos == endpos || endpos > p))
		    || (p + snip->count == len && startpos == len));
	
	if (hilite && (snip->flags & wxSNIP_NEWLINE))
	  /* End of line: */
	  hilite = ((startpos != p + snip->count)
		    || (startpos == endpos && posateol)
		    || (startpos != endpos && startpos < p + snip->count));
	if (hilite && PTREQ(snip, first))
	  /* Beginning of line: */
	  hilite = ((endpos != p)
		    || (startpos == endpos && !posateol)
		    || (startpos != endpos && endpos > p));
      } else
	hilite = FALSE;

      if (hilite) {
	bottom = down + h;

	if (startpos <= p)
	  hxs = x;
	else
	  hxs = x + snip->PartialOffset(dc, x, ycounter, startpos - p);
	
	if (endpos >= p + snip->count) {
	  if (snip->flags & wxSNIP_NEWLINE) {
	    if (endpos == startpos)
	      hxe = hxs;
	    else {
	      hxe = rightx;
	      bottom = ycounter + line->h;
	    }
	  } else
	    hxe = x + w;
	} else
	  hxe = x + snip->PartialOffset(dc, x, ycounter, endpos - p);
	
	if (!hiliteSome) {
	  hsxs = hxs;
	  hsxe = hxe;
	  hsys = down;
	  hsye = bottom;
	  hiliteSome = TRUE;
	} else {
	  hsxe = hxe;
	  if (down < hsys)
	    hsys = down;
	  if (bottom > hsye)
	    hsye = bottom;
	}
      }

      x += w;
      
      p += snip->count;
    }

    if (hiliteSome) {
      if (hsxe != hsxs) {
	  savePen = dc->GetPen();
	  saveBrush = dc->GetBrush();
	  dc->SetPen(outlinePen);
	  dc->SetBrush(outlineBrush);
	  
	  dc->SetLogicalFunction(wxXOR);
	  dc->DrawRectangle(hsxs + dx, hsys + dy, 
			    hsxe - hsxs + 1 + DELTA, hsye - hsys + 1 + DELTA);
	  dc->SetLogicalFunction(wxCOPY);
	  
	  dc->SetPen(savePen);
	  dc->SetBrush(saveBrush);
	} else {
	  savePen = dc->GetPen();
	  dc->SetPen(caretPen);
	  dc->SetLogicalFunction(wxXOR);
	  dc->DrawLine(hsxs + dx, hsys + dy, 
		       hsxs + dx, hsye + dy + 1 + DELTA);
	  dc->SetLogicalFunction(wxCOPY);
	  dc->SetPen(savePen);

	  caretLocationX = hsxs;
	  caretLocationT = hsys;
	  caretLocationB = hsye + 1 + DELTA;
	  caretOn = TRUE;
	}
    }

    pcounter += line->len;
    ycounter += line->h;
  }

  styleList->BasicStyle()->SwitchTo(dc, oldStyle);

  if (show_caret && !caretSnip)
    if (!line && extraLine)
      if (!posateol && startpos == len && endpos == startpos) {
	float y;
	y = ycounter;

	savePen = dc->GetPen();
	dc->SetPen(caretPen);
	dc->SetLogicalFunction(wxXOR);
	dc->DrawLine(dx, y + dy, dx, y + extraLineH + dy + DELTA);
	dc->SetLogicalFunction(wxCOPY);
	dc->SetPen(savePen);

	caretLocationX = 0;
	caretLocationT = y;
	caretLocationB = y + extraLineH + DELTA;
	caretOn = TRUE;
      }
 
paint_done:

  OnPaint(FALSE, dc, leftx, starty, rightx, endy, dx, dy, show_caret);

  dc->EndDrawing();


  locked = FALSE;
}

/* This one notifies the administrator that we need to be updated. */
void wxMediaEdit::Redraw()
{
  float w, h;
  float top, bottom, height, width, left, right;
  float x, y, origx, origy;
  float fy, fx;
  Bool oneline;
  wxDC *dc;

  if (locked || !admin)
    return;

  dc = admin->GetDC(&x, &y);

  if (!dc)
    return;

  origx = x;
  origy = y;

  RecalcLines(dc);

  if (delayedscroll != -1) {
    if (ScrollToPosition(delayedscroll, delayedscrollateol, FALSE,
			 delayedscrollend, delayedscrollbias))
      refreshAll = TRUE;
  } else if (delayedscrollbox) {
    delayedscrollbox = FALSE;
    if (ScrollTo(delayedscrollsnip, delayedscrollX, delayedscrollY,
		 delayedscrollW, delayedscrollH, FALSE, delayedscrollbias))
      refreshAll = TRUE;
  }

  admin->GetDC(&x, &y);
  
  if (x != origx || y != origy)
    refreshAll = TRUE;

  admin->GetView(NULL, NULL, &w, &h, TRUE);

  top = y;
  bottom = y + h;
  left = x;
  right = left + w;

  /* Figure out the minimal refresh area. The refresh area may be
     determined by character position ranges, box coordinates, or
     both. If neither is specified, we have to assume that everything
     needs to be refreshed. */
  if (!refreshAll && (!refreshUnset || !refreshBoxUnset)) {
    if (!refreshUnset) {
      top = y;
      if (refreshStart > -1 && refreshEnd > -1) {
	// oneline = PositionLine(refreshStart) == PositionLine(refreshEnd);
	oneline = FALSE; // b/c it doesn't seem to help
      } else
	oneline = FALSE;
      if (refreshStart > -1) {
	PositionLocation(refreshStart, 
			 oneline ? (float *)&fx : (float *)NULL, (float *)&fy, 
			 TRUE, TRUE, TRUE);
	if (fy > top)
	  top = (long)fy;
	if (oneline && fx > left)
	  left = fx;
      }
      bottom = y + h;
      if (refreshEnd > -1) {
	PositionLocation(refreshEnd, 
			 oneline ? (float *)&fx : (float *)NULL, (float *)&fy, 
			 FALSE, FALSE, TRUE);
	if (fy < bottom)
	  bottom = (long)fy;
	if (oneline && fx + CURSOR_WIDTH < right)
	  right = fx + CURSOR_WIDTH;
      }

      if (!refreshBoxUnset) {
	if (refreshT < top)
	  top = refreshT;
	if (refreshB > bottom)
	  bottom = refreshB;
      }
    } else {
      if (refreshL > left)
	left = refreshL;
      if (refreshR < right)
	right = refreshR;
      if (refreshT > top)
	top = refreshT;
      if (refreshB < bottom)
	bottom = refreshB;
    }
  } else if (!refreshAll)
    return; /* Nothing needs to be updated! */

  refreshUnset = refreshBoxUnset = TRUE;
  refreshAll = FALSE;

  height = bottom - top;
  width = right - left;

  admin->NeedsUpdate(left, top, width, height);
}

/* This one is called by the administrator: */
void wxMediaEdit::Refresh(float left, float top, float width, float height,
			  Bool show_caret)
{
  float x, y, bottom, right, ddx, ddy;
  Bool ps;
  wxDC *dc;

  if (!width || !height)
    return;

  dc = admin->GetDC(&x, &y);
  if (!dc)
    return;

  caretLocationX = -1;
  caretOn = FALSE;

  if (height > bmHeight || width > bmWidth) {
    wxBitmap *oldbm = bitmap;

    bmWidth = ROUND(width);
    bmHeight = ROUND(height);

    bitmap = new wxBitmap(bmWidth, bmHeight);

    offscreen->SelectObject(NULL);
    if (oldbm)
      delete oldbm;
  
    if (bitmap->Ok() && offscreen->Ok())
      offscreen->SelectObject(bitmap);

    drawCachedInBitmap = FALSE;
  } 

  bottom = top + height;
  right = left + width;

  ps = wxSubType(dc->__type, wxTYPE_DC_POSTSCRIPT);

  if (bitmap && bitmap->Ok() && offscreen->Ok() && !ps) {
    /* Need to make sure that difference between coordinates is
       integral; otherwise, roundoff error could affect drawing */
    ddx = (left - x) - (long)(left - x);
    if (ddx < 0)
      ddx = 1 + ddx;
    left -= ddx;
    width += ddx;
    ddy = (top - y) - (long)(top - y);
    if (ddy < 0)
      ddy = 1 + ddy;
    top -= ddy;
    height += ddy;

    if (!drawCachedInBitmap || (top != lastDrawT) || (bottom != lastDrawB)
	|| (left != lastDrawL) || (right != lastDrawR)
	|| (lastDrawCaret != show_caret)) {
      Redraw(offscreen, top, bottom, left, right, -top, -left, show_caret);
      lastDrawL = left;
      lastDrawT = top;
      lastDrawR = right;
      lastDrawB = bottom;
      lastDrawCaret = show_caret;
      drawCachedInBitmap = TRUE;
    }
    dc->Blit(left - x, top - y, width, height, offscreen, 0, 0, wxCOPY);
  } else {
    if (ps)
      skipBox = TRUE;
    Redraw(dc, top, bottom, left, right, -y, -x, show_caret);
    if (ps)
      skipBox = FALSE;
  }

  if (changed) {
    changed = FALSE;
    OnChange();
  }
}

/* This one is used internally to delay refreshes: */
void wxMediaEdit::NeedRefresh(long start, long end)
{
  if (refreshUnset) {
    refreshStart = start;
    refreshEnd = end;
    refreshUnset = FALSE;
  } else {
    if (start < refreshStart)
      refreshStart = start;
    if (end == -1)
      refreshEnd = -1;
    else if (refreshEnd != -1 && end > refreshEnd)
      refreshEnd = end;
  }

  drawCachedInBitmap = FALSE;

  if (!delayRefresh)
    Redraw();
  else if (admin && !admin->standard)
    admin->Resized(FALSE);
}

void wxMediaEdit::RefreshByLineDemand(void)
{
  if (!graphicMaybeInvalid)
    graphicMaybeInvalid = TRUE;

  if (!delayRefresh)
    Redraw();
  else if (admin && !admin->standard)
    admin->Resized(FALSE);
}

void wxMediaEdit::NeedCaretRefresh(void)
{
  if (!admin || (admin->standard <= 0) || delayRefresh 
      || startpos != endpos || flash)
    NeedRefresh(startpos, endpos);
  else if (ownCaret)
    CaretOn();
  else
    CaretOff();
}

void wxMediaEdit::CalcCaretLocation(void)
{
  if (caretLocationX < 0) {
    PositionLocation(startpos, &caretLocationX, &caretLocationT,
		     TRUE, posateol, FALSE);
    PositionLocation(startpos, NULL, &caretLocationB,
		     FALSE, posateol, FALSE);
  }
}

Bool wxMediaEdit::CaretOff(void)
{
  wxDC *dc;
  float dx, dy, x, y, w, h, X, T, B;

  if (!CheckRecalc())
    return FALSE;

  if (refreshAll || !refreshUnset || !refreshBoxUnset || (delayedscroll!=-1)) {
    Redraw();
    return FALSE;
  }

  dc = admin->GetDC(&dx, &dy);

  if (!dc)
    return FALSE;

  admin->GetView(&x, &y, &w, &h);

  CalcCaretLocation();

  X = caretLocationX;
  T = caretLocationT;
  B = caretLocationB;

  if (B < y)
    return TRUE;
  if (T >= y + h)
    return TRUE;
  if (X < x || X >= x + w)
    return TRUE;

  if (T < y)
    T = y;
  if (B > y + h)
    B = y + h - 1;

  if (!caretPen)
    caretPen = new wxPen("BLACK", 0, wxSOLID);

  dc->SetPen(caretPen);
  dc->SetLogicalFunction(wxXOR);
  dc->DrawLine(X - dx, T - dy, X - dx, B - dy);
  dc->SetLogicalFunction(wxCOPY);

  drawCachedInBitmap = FALSE;

  caretOn = FALSE;

  return TRUE;
}

void wxMediaEdit::CaretOn(void)
{
  if (CaretOff())
    caretOn = TRUE;
}

/* 8.5" x 11" Paper, 0.5" Margin; usually not used */
static long page_width = 612, page_height = 792;
/* Printing margins: */
static long h_margin = 36, v_margin = 36;

void wxGetMediaPrintMargin(long *hm, long *vm)
{
  if (hm)
    *hm = h_margin;
  if (vm)
    *vm = v_margin;
}

void wxSetMediaPrintMargin(long hm, long vm)
{
  if (hm > -1)
    h_margin = hm;
  if (vm > -1)
    v_margin = vm;
}

void wxMediaEdit::PrintToDC(wxDC *dc)
{
  float H, W, FH, FW, y, h;
  int i;
  wxMediaLine *line;

  CheckRecalc();

  SizeCacheInvalid();
  RecalcLines(dc, TRUE);

  dc->GetSize(&W, &H);

  if (!W || !H) {
    W = page_width;
    H = page_height;
  }
  
  if (wxGetPrinterOrientation() != PS_PORTRAIT)  {
    float tmp;

    tmp = H;
    H = W;
    W = tmp;
  }

  FH = H;
  FW = W;

  H -= (2 * v_margin);
  W -= (2 * h_margin);

  y = 0;
  line = firstLine;
  for (i = 0; i < numValidLines; ) {
    h = 0;
    while (!h || ((i < numValidLines) && (line->h < H - h))) {
      h += line->h;
      i++;
      line = line->next;
    }

    dc->StartPage();

    /* Establish page size: */
    dc->DrawLine(0, 0, 0, 0);
    dc->DrawLine(FW, FH, FW, FH);

    skipBox = TRUE;
    Redraw(dc, y, y + h, 0, W, -y + v_margin, h_margin, FALSE);
    skipBox = FALSE;

    dc->EndPage();

    y += h + 1;
  }

  SizeCacheInvalid();  
}
